"""
activity_log.py - Contains the ActivityLog class, which represents
a single activity log entry.
"""

from dataclasses import dataclass, field, asdict
from typing import Union, Optional
from datetime import datetime
from ipaddress import ip_address, IPv4Address, IPv6Address


@dataclass(order=True)
class ActivityLog:
    """
    Represents a single activity log entry.
    """

    Date: Union[str, datetime] = field(default=None)
    Action: Optional[str] = field(default=None, compare=False)
    Module: Optional[str] = field(default=None, compare=False)
    Details: Optional[str] = field(default=None, compare=False)
    User_Name: Optional[str] = field(default=None, compare=False)
    User_Role: Optional[str] = field(default=None, compare=False)
    User_IP: Optional[Union[str, ip_address]] = field(default=None, compare=False)

    def __post_init__(self):
        if isinstance(self.Date, str):
            self.Date = datetime.fromisoformat(self.Date)

        if self.User_IP and not type(self.User_IP) in [IPv4Address, IPv6Address]:
            try:
                self.User_IP = ip_address(self.User_IP)
            except ValueError:
                if self.User_IP == "N/A":
                    self.User_IP = None
                else:
                    raise ValueError(f"Invalid IP address string: {self.User_IP}")

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def __str__(self):
        return f"{self.Date} - {self.Action} - {self.Module} - {self.Details} - {self.User_Name} - {self.User_Role} - {self.User_IP}"

    def get_action(self):
        return self.Action

    def get_module(self):
        return self.Module

    def from_dict(data: dict):
        return ActivityLog(**data)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()



"""
Global AssetView API (GAV) module

This module contains ways to interact with the Qualys Global AssetView API (GAV). Valid endpoints are defined in the CALL_SCHEMA dictionary.

GAV QQL Syntax help: https://docs.qualys.com/en/gav/2.18.0.0/search/how_to_search.htm
"""

from .count_assets import count_assets
from .get_all_assets import get_all_assets
from .get_asset import get_asset
from .query_assets import query_assets
from .uber import GAVUber



"""
Contains function to pull details on all cloud controls Qualys checks for.
"""

from typing import Union

from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.Controls import Control


def get_control_metadata(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
):
    """
    Get controls Qualys checks a cloud provider for.

    Args:
        auth (BasicAuth): The authentication object.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        filter (str): Filter controls by providing a query. Queryable fields are:
        - control.name
        - resource.type
        - service.type
        - cid (supports control for IaC evaluations)
        - provider
        - control.criticality
        - control.type
        - policy.name
        - createdDate
        - modifiedDate
        - isCustomizable
        - qflow.name
        - qflow.id

        Example: filter="provider:AWS"

        pageNo (int): The ordered page to start retrieving controls from, or if page_count is 1, the page to retrieve.
        pageSize (int): The number of controls to get per page.

    Returns:
        BaseList[Control]: The response from the API as a BaseList of Control objects.
    """

    responses = BaseList()
    currentPage = 0

    if kwargs.get("filter"):
        # Make sure the query key is valid
        valid_controls = [
            "control.name",
            "resource.type",
            "service.type",
            "cid",
            "provider",
            "control.criticality",
            "control.type",
            "policy.name",
            "createdDate",
            "modifiedDate",
            "isCustomizable",
            "qflow.name",
            "qflow.id",
        ]

        if not kwargs["filter"].split(":")[0] in valid_controls:
            raise QualysAPIError(
                f"Invalid filter key. Valid keys: {', '.join(valid_controls)}"
            )

        # If provider, resource.type, control.criticality,
        # service.type, control.type is used in the filter, uppercase the value.
        if (
            "provider" in kwargs["filter"]
            or "resource.type" in kwargs["filter"]
            or "control.criticality" in kwargs["filter"]
            or "service.type" in kwargs["filter"]
            or "control.type" in kwargs["filter"]
        ):
            kwargs[
                "filter"
            ] = f"{kwargs['filter'].split(':')[0]}:{kwargs['filter'].split(':')[1].upper()}"

    while True:
        # Set the current page number and page size in kwargs
        kwargs["pageNo"] = currentPage

        # Make the API request to retrieve the connectors
        response = call_api(
            auth=auth,
            module="cloudview",
            endpoint="get_control_metadata",
            params=kwargs,
            headers={"accept": "application/json"},
        )

        if response.status_code != 200:
            raise QualysAPIError(
                f"Error retrieving controls. Status code: {response.status_code}. Requests reporting: {response.text}"
            )

        j = response.json()

        if len(j["control"]) == 0:
            print("No controls found.")
            break

        # Iterate through the records in the response and create Connector objects
        for record in j["control"]:
            responses.append(Control.from_dict(record))

        # Print a message indicating the current page was retrieved successfully
        print(f"Page {currentPage+1} of controls retrieved successfully.")
        currentPage += 1

        # Break the loop if all pages are retrieved or the requested number of pages are retrieved
        if (
            page_count != "all" and (currentPage + 1 > page_count)
        ) or "warning" not in j.keys():
            break

    # Print a message indicating all pages have been retrieved
    print(f"All pages complete. {str(len(responses))} control records retrieved.")

    return responses



"""
uber.py - an Uber class that allows for easy access to all GAV API endpoints,
plus some extra functionality such as exporting API results to files/SQL databases.
"""

from typing import Union, Literal

from . import *
from ..base.call_schema import CALL_SCHEMA
from ..auth import *
from ..exceptions.Exceptions import *


class GAVUber:
    """
    The Uber class is a class that allows for easy access to all GAV API endpoints.
    """

    def __init__(self, auth: Union[TokenAuth, BasicAuth]):
        """
        Initialize the Uber class.
        """
        self.NON_ENDPOINTS = ["url_type"]  # may expand over time.. your move, Qualys.
        self.auth = auth
        self.valid_endpoints = [
            i for i in CALL_SCHEMA["gav"].keys() if i not in self.NON_ENDPOINTS
        ]  # grab actual endpoints

        # check if the auth type is valid:
        if self.auth.auth_type not in ["token", "basic"]:
            raise InvalidAuthTypeError(
                f"Invalid auth type: {self.auth.auth_type}. Acceptable auth types are: 'token' or 'basic'."
            )

    def __str__(self) -> str:
        return f"Uber class for GAV API with auth type: {self.auth.auth_type}."

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def get(
        self,
        endpoint: Literal[
            "count_assets", "get_all_assets", "get_asset", "query_assets"
        ],
        **kwargs,
    ):
        """
        Call the appropriate endpoint based on passed endpoint.

        Params:
            endpoint (str): The endpoint to call. Must be one of the keys in CALL_SCHEMA["gav"].
            **kwargs: The keyword arguments to pass to the endpoint.

        Returns:
            The response from the API call.
        """

        match endpoint:
            case "count_assets":
                return count_assets(self.auth, **kwargs)
            case "get_all_assets":
                return get_all_assets(self.auth, **kwargs)
            case "get_asset":
                return get_asset(self.auth, **kwargs)
            case "query_assets":
                return query_assets(self.auth, **kwargs)
            case _:
                raise InvalidEndpointError(
                    f"Invalid endpoint: {endpoint}. Acceptable endpoints are: {self.valid_endpoints}."
                )



"""
Helper functions for multithreading get_hld and get_host_list functions.
"""

from queue import Queue, Empty
from typing import Union, List, Literal
from threading import current_thread, Lock
from urllib.parse import urlparse, parse_qs
from os import cpu_count

from ..data_classes.hosts import VMDRID, VMDRHost
from ...exceptions import QualysAPIError
from ...base.call_api import call_api
from ...base.xml_parser import xml_parser
from ...auth.basic import BasicAuth
from ...base.base_list import BaseList

LOCK = Lock()


def prepare_args(
    auth: BasicAuth,
    chunk_size: int,
    threads: int,
    page_count: Union[int, "all"],
    chunk_count: Union[int, "all"],
    ids: str = None,
) -> dict:
    """
    Performs necessary checks for thread count, CPU count, rate
    limits.

    Args:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        chunk_size (int): The size of each chunk.
        threads (int): The number of threads.
        page_count (Union[int, 'all']): The number of pages to retrieve.
        chunk_count (Union[int, 'all']): The number of chunks to retrieve.
        ids (str): A comma-separated string of host IDs to use. If specified, this will be used instead of pulling the full set.

    Returns:
        None

    Raises:
        ValueError: If threads, chunk_size, page_count, or chunk_count are less than 1.
    """
    # Ensure that threads, chunk_size, and page_count (if not 'all') are all integers above 0
    if any(
        [
            threads < 1,
            chunk_size < 1,
            (page_count != "all" and page_count < 1),
            (chunk_count != "all" and chunk_count < 1),
        ]
    ):
        raise ValueError(
            "threads, chunk_size, page_count (if not 'all') and chunk_count (if not 'all') must all be integers above 0."
        )

    # Make sure the user hasn't set threads to more than the cpu count
    if threads > cpu_count():
        with LOCK:
            print(
                f"Warning: The number of threads ({threads}) is greater than the number of CPUs ({cpu_count()}). This may cause performance issues."
            )

    # Second, get concurrency rate limit from auth object. NOTE: eventually, auth class should have an attribute for this.
    rl = auth.get_ratelimit()
    if threads > rl["X-Concurrency-Limit-Limit"]:
        with LOCK:
            print(
                f"Warning: The number of threads ({threads}) is greater than the concurrency rate limit ({rl['X-Concurrency-Limit-Limit']}). Setting threads to {rl['X-Concurrency-Limit-Limit']}."
            )
        threads = rl["X-Concurrency-Limit-Limit"]

    (
        print(f"Pulling/creating queue for full ID list...")
        if not ids
        else print(f"Pulling/creating queue for user-specified IDs: {ids}...")
    )


def normalize_id_list(id_list):
    """
    normalize_id_list - formats the kwarg ids, if it is passed.

    Since the API accepts multiple values, either as comma-separated or as a range, this function
    normalizes the input to a list of integers that satisfy the API requirements.
    """
    id_list = id_list.split(",")
    new_list = []
    for i in id_list:
        if "-" in i:
            # if there is a hyphen, split by hyphen and create a range
            new_list.extend(range(int(i.split("-")[0]), int(i.split("-")[1]) + 1))
        else:
            # if there is no hyphen, just append the integer
            new_list.append(int(i))
    # Sort the list to ensure the range is correct
    new_list.sort()
    return new_list


def pull_id_set(auth: BasicAuth, ids: str = None) -> BaseList[VMDRID]:
    """
    pull_id_set - pull a set of host IDs from the VMDR API.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        ids (str): A comma-separated string of host IDs to use. If specified, this will be used instead of pulling the full set.

    Returns:
        List[int]: A BaseList of host IDs as VMDRIDs.
    """

    if ids:
        # User has specified specific IDs to pull:
        res = get_host_list_backend(auth, details=None, truncation_limit=0, ids=ids)
    else:
        # User has not specified specific IDs to pull, so pull all:
        res = get_host_list_backend(auth, details=None, truncation_limit=0)

    if not res:
        raise QualysAPIError("No IDs returned from API!")

    return [str(i) for i in res]


def create_id_queue(
    auth: BasicAuth, chunk_size: int = 100, ids: str = None
) -> BaseList:
    """
    create_id_queue - create a queue of host IDs to pull. the queue contains chunks (lists) of chunk_size
    length with host IDs.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        chunk_size (int): The size of each chunk. Defaults to 100.
        ids (str): A comma-separated string of host IDs to use. If specified, this will be used instead of pulling the full set.

    Returns:
        Queue: A queue of host IDs to pull.
    """

    if ids:
        id_list = pull_id_set(auth, ids=ids)
    else:
        id_list = pull_id_set(auth)

    if not id_list:
        raise QualysAPIError("No IDs returned from API.")
    with LOCK:
        print(f"ID set pulled. Total IDs: {len(id_list)}")

    id_queue = Queue()

    for i in range(0, len(id_list), chunk_size):
        id_queue.put(id_list[i : i + chunk_size])

    singular_chunk = True if id_queue.qsize() == 1 else False

    # If there is only 1 chunk, split it up where each new
    # chunk only contains 1 ID. This is to ensure that each
    # thread has something to pull.
    if singular_chunk:
        with LOCK:
            print(
                "Only 1 chunk of IDs. Splitting it up into chunks of 1 ID each to ensure threads have work to do."
            )
        new_queue = Queue()
        for i in range(0, len(id_list)):
            new_queue.put(id_list[i : i + 1])
        id_queue = new_queue

    with LOCK:
        print(
            f"Queue created with {id_queue.qsize()} chunks of ~{chunk_size} IDs each."
        )

    return id_queue


def hld_backend(
    auth: BasicAuth,
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> List:
    """
    hld_backend - get a list of hosts and their QID detections.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        **kwargs: Additional keyword arguments to pass to the API. See below.

    Kwargs:

        action (Optional[str]) #The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]) #Whether to echo the request. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        show_asset_id (Optional[bool]) #Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        include_vuln_type (Optional[Literal["confirmed", "potential"]]) #The type of vulnerability to include. If not specified, both types are included.

        DETECTION FILTERS:
        show_results (Optional[bool]) #Whether to show the results. Default is True. Ends up being passed to the API as 0 or 1. WARNING: this SDK overwrites any value you pass with '1'. It is just recognized as valid for the sake of completeness.
        arf_kernel_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns for Linux kernels. 0 = don't filter, 1 = exclude kernel vulns that are not exploitable, 2 = only include kernel related vulns that are not exploitable, 3 = only include exploitable kernel vulns, 4 = only include kernel vulns. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        arf_service_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns found on running or nonrunning ports/services. 0 = don't filter, 1 = exclude service related vulns that are exploitable, 2 = only include service vulns that are exploitable, 3 = only include service vulns that are not exploitable, 4 = only include service vulns. If specified, results are in a host's <AFFECT_RUNNING_SERVICE> tag.
        arf_config_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns that can be vulnerable based on host config. 0 = don't filter, 1 = exclude vulns that are exploitable due to host config, 2 = only include config related vulns that are exploitable, 3 = only include config related vulns that are not exploitable, 4 = only include config related vulns. If specified, results are in a host's <AFFECT_EXPLOITABLE_CONFIG> tag.
        active_kernels_only (Optional[Literal[0,1,2,3]]) #Specify vulns related to running or non-running kernels. 0 = don't filter, 1 = exclude non-running kernels, 2 = only include vulns on non-running kernels, 3 = only include vulns with running kernels. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        output_format (Optional[Literal["XML", "CSV"]]) #The format of the output. Default is 'XML'. WARNING: this SDK will overwrite any value you pass with 'XML'. It is just recognized as valid for the sake of completeness.
        supress_duplicated_data_from_csv (Optional[bool]) #Whether to suppress duplicated data from CSV. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        truncation_limit (Optional[int]) #The truncation limit for a page. Default is 1000.
        max_days_since_detection_updated (Optional[int]) #The maximum number of days since the detection was last updated. For detections that have never changed, the value is applied to the last detection date.
        detection_updated_since (Optional[str]) #The date and time since the detection was updated.
        detection_updated_before (Optional[str]) #The date and time before the detection was updated.
        detection_processed_before (Optional[str]) #The date and time before the detection was processed.
        detection_processed_after (Optional[str]) #The date and time after the detection was processed.
        detection_last_tested_since (Optional[str]) #The date and time since the detection was last tested.
        detection_last_tested_since_days (Optional[int]) #The number of days since the detection was last tested.
        detection_last_tested_before (Optional[str]) #The date and time before the detection was last tested.
        detection_last_tested_before_days (Optional[int]) #The number of days before the detection was last tested.
        include_ignored (Optional[bool]) #Whether to include ignored detections. Default is False. Ends up being passed to the API as 0 or 1.
        include_disabled (Optional[bool]) #Whether to include disabled detections. Default is False. Ends up being passed to the API as 0 or 1.

        HOST FILTERS:
        ids (Optional[str]) #A comma-separated string of host IDs to include.
        id_min (Optional[Union[int,str]]) #The minimum host ID to include.
        id_max (Optional[Union[int,str]]) #The maximum host ID to include.
        ips (Optional[str]) #The IP address of the host to include. Can be a comma-separated string, and also supports ranges with a hyphen: 10.0.0.0-10.0.0.255.
        ipv6 (Optional[str]) #The IPv6 address of the host to include. Can be a comma-separated string. Does not support ranges.
        ag_ids (Optional[str]) #Show only hosts belonging to the specified asset group IDs. Can be a comma-separated string, and also supports ranges with a hyphen: 1-5.
        ag_titles (Optional[str]) #Show only hosts belonging to the specified asset group titles. Can be a comma-separated string.
        network_ids (Optional[str]) #Show only hosts belonging to the specified network IDs. Can be a comma-separated string.
        network_names (Optional[str]) #displays the name of the network corresponding to the network ID.
        vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        no_vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        max_days_since_last_vm_scan (Optional[int]) #The maximum number of days since the last VM scan.
        vm_processed_before (Optional[str]) #The date and time before the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_processed_after (Optional[str]) #The date and time after the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_before (Optional[str]) #The date and time before the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_after (Optional[str]) #The date and time after the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_before (Optional[str]) #The date and time before the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_after (Optional[str]) #The date and time after the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        status (Optional[Literal["New", "Active", "Fixed", "Re-Opened"]]) #The status of the detection.
        compliance_enabled (Optional[bool]) #Whether compliance is enabled. Default is False. Ends up being passed to the API as 0 or 1.
        os_pattern (Optional[str]) #PCRE Regex to match operating systems.

        QID FILTERS:
        qids (Optional[str]) #A comma-separated string of QIDs to include.
        severities (Optional[str]) #A comma-separated string of severities to include. Can also be a hyphenated range, i.e. '2-4'.
        filter_superseded_qids (Optional[bool]) #Whether to filter superseded QIDs. Default is False. Ends up being passed to the API as 0 or 1.
        include_search_list_titles (Optional[str]) #A comma-separated string of search list titles to include.
        exclude_search_list_titles (Optional[str]) #A comma-separated string of search list titles to exclude.
        include_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to include.
        exclude_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to exclude.

        ASSET TAG FILTERS:
        use_tags (Optional[bool]) #Whether to use tags. Default is False. Ends up being passed to the API as 0 or 1.
        tag_set_by (Optional[Literal['id','name']]) #When filtering on tags, whether to filter by tag ID or tag name.
        tag_include_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_exclude_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_set_include (Optional[str]) #A comma-separated string of tag IDs or names to include.
        tag_set_exclude (Optional[str]) #A comma-separated string of tag IDs or names to exclude.
        show_tags (Optional[bool]) #Whether to show tags. Default is False. Ends up being passed to the API as 0 or 1.

        QDS FILTERS:
        show_qds (Optional[bool]) #Whether to show QDS. Default is False. Ends up being passed to the API as 0 or 1.
        qds_min (Optional[int]) #The minimum QDS to include.
        qds_max (Optional[int]) #The maximum QDS to include.
        show_qds_factors (Optional[bool]) #Whether to show QDS factors. Default is False. Ends up being passed to the API as 0 or 1.

        EC2/AZURE METADATA FILTERS:
        host_metadata (Optional[Literal['all,'ec2', 'azure']]) #The type of metadata to include. Default is 'all'.
        host_metadata_fields (Optional[str]) #A comma-separated string of metadata fields to include. Use carefully.
        show_cloud_tags (Optional[bool]) #Whether to show cloud tags. Default is False. Ends up being passed to the API as 0 or 1.
        cloud_tag_fields (Optional[str]) #A comma-separated string of cloud tag fields to include. Use carefully.

    Returns:
        List: A list of VMDRHost objects, with their DETECTIONS attribute populated.
    """

    # Set the kwargs
    kwargs["action"] = "list"
    kwargs["echo_request"] = 0
    kwargs["show_results"] = 1
    kwargs["output_format"] = "XML"

    responses = BaseList()
    pulled = 0

    while True:
        with LOCK:
            print(
                f"{current_thread().name} - Pulling page {pulled+1} for ids {kwargs.get('ids')}. KWARGS: {kwargs}"
            )

        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_hld",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )

        if response.status_code != 200:
            with LOCK:
                print(f"{current_thread().name} - No data returned on page {pulled}")
            pulled += 1
            if pulled != "all":
                if pulled == page_count:
                    with LOCK:
                        print(f"{current_thread().name} - Pulled all pages.")
                    break
                else:
                    continue

        # cleaned = remove_problem_characters(response.text)
        xml = xml_parser(response.content)

        # check if there is no host list
        if "HOST_LIST" not in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]:
            with LOCK:
                print(f"{current_thread().name} - No host list returned.")
        else:
            # check if ["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"] is a list of dictionaries
            # or just a dictionary. if it is just one, put it inside a list
            if not isinstance(
                xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"],
                list,
            ):
                xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                    "HOST"
                ] = [
                    xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                        "HOST"
                    ]
                ]

            for host in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                "HOST"
            ]:
                host_obj = VMDRHost.from_dict(host)
                responses.append(host_obj)

        pulled += 1
        if page_count != "all":
            if pulled == page_count:
                break

        if "WARNING" in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]:
            if "URL" in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["WARNING"]:
                # get the id_min parameter from the URL to pass into kwargs:
                params = parse_qs(
                    urlparse(
                        xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["WARNING"][
                            "URL"
                        ]
                    ).query
                )
                with LOCK:
                    print(
                        f"{current_thread().name} (get_hld) - Pagination detected. Pulling next page with id_min: {params['id_min'][0]}"
                    )
                kwargs["id_min"] = params["id_min"][0]

            else:
                break
        else:
            break

    return responses


def thread_worker(
    auth: BasicAuth,
    id_queue: Queue,
    responses: BaseList,
    page_count: Union[int, "all"],
    chunk_count: Union[int, "all"],
    endpoint_called: Literal["get_hld", "get_host_list"],
    kwargs,
):
    """
    thread_worker - the worker function for get_hld/hld_backend functions.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        id_queue (Queue): The queue of host IDs to pull.
        responses (BaseList): The list of responses to append to.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        chunk_count (Union[int, "all"]): The number of chunks to retrieve. Defaults to "all".
        endpoint_called (Union['get_hld', 'get_host_list']): The function that was called.
        **kwargs: Additional keyword arguments to pass to the API. See get_hld() for details.
    """

    while True:
        pages_pulled = 0
        chunks_pulled = 0
        try:
            ids = (
                id_queue.get_nowait()
            )  # nowait allows us to check if the queue is empty without blocking
        except Empty:
            with LOCK:
                print(f"{current_thread().name} - Queue is empty. Terminating thread.")
            break

        if not ids:
            with LOCK:
                print(f"{current_thread().name} - No IDs to pull. Terminating thread.")
            break

        if len(ids) != 1:
            kwargs["ids"] = f"{ids[0]}-{ids[-1]}"
        else:
            kwargs["ids"] = ids[0]

        if endpoint_called == "get_hld":
            responses.extend(hld_backend(auth, page_count=page_count, **kwargs))
        elif endpoint_called == "get_host_list":
            responses.extend(
                get_host_list_backend(auth, page_count=page_count, **kwargs)
            )
        else:
            raise ValueError(
                "endpoint_called must be either 'get_hld' or 'get_host_list'."
            )
        id_queue.task_done()
        with LOCK:
            print(f"{current_thread().name} ({endpoint_called}) - Chunk complete.")
        pages_pulled += 1
        chunks_pulled += 1
        # check if the queue is empty, or if the threads are done (via pulled var)
        if id_queue.empty():
            with LOCK:
                print(
                    f"{current_thread().name} ({endpoint_called}) - Queue is empty. Terminating thread."
                )
            break
        if pages_pulled == page_count:
            with LOCK:
                print(
                    f"{current_thread().name} - Thread has pulled all pages. Terminating thread."
                )
            break
        if chunks_pulled == chunk_count:
            with LOCK:
                print(
                    f"{current_thread().name} - Thread has pulled all chunks. Terminating thread."
                )
            break


def get_host_list_backend(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> list:
    """
    Get the host list from the VMDR API.
    For a full list of parameters, see the Qualys API documentation: [Qualys API VMPc User Guide](https://cdn2.qualys.com/docs/qualys-api-vmpc-user-guide.pdf)

    Parameters:
        auth (BasicAuth): The authentication object.
        page_count (Union[int, "all"]): The number of pages to get. If "all", get all pages. Defaults to "all".

    :Kwargs:

        action (Optional[str]): The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]): Whether to show the request. Default is 'False'. ends up being passed to API as 0 or 1.
        show_asset_id (Optional[bool]): Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        details (Optional[Union[Literal["Basic", "Basic/AGs", "All", "All/AGs", "None"], None]]): The level of detail to return. Default is 'Basic'. Basic includes host ID, IP, tracking method, DNS, netBIOS, and OS. Basic/AGs includes basic host information plus asset group information. Asset group information includes the asset group ID and title. All shows all basic host information plus last vulnerability and compliance scan dates. All/AGs includes all information plus asset group information: group ID and title. None shows only the host ID (or asset ID if show_asset_id is set to 1 (True)).
        os_pattern (Optional[str]): The OS regex pattern to search for. It follows the PCRE standard, and must be URL-encoded. To match an empty string, use '%5E%24'.
        truncation_limit (Optional[int]): The maximum number of characters to return for each field. Default is 1000, and can be set to 0 for no truncation all the way up to 1,000,000. Past 1M, the API will return an error.
        ips (Optional[str]): Only show assets with certain IP addresses/ranges. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 10.0.0.1-10.0.0.255
        ipv6 (Optional[str]): Only show assets with certain IPv6 addresses/ranges. Multiple values specified as a comma-separated string.
        ag_ids (Optional[str]): Only show assets in certain asset group IDs. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 1-5. A valid ID is required or the API will return an error.
        ag_titles (Optional[str]): Only show assets with specified string in the asset group title. Multiple values specified as a comma-separated string.
        ids (Optional[str]): Only show assets with certain asset IDs. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 1-5. A valid ID is required or the API will return an error.
        id_min (Optional[int]): Only show assets with asset IDs greater than or equal to this value.
        id_max (Optional[int]): Only show assets with asset IDs less than or equal to this value.
        network_ids (Optional[str]): Valid only when the Network Support feature is enabled for the users account). Restrict the request to certain custom network IDs. Multiple network IDs are comma separated.
        compliance_enabled (Optional[bool]): Only show assets with compliance enabled. Default is 'False'. ends up being passed to API as 0 or 1.

        DATE PARAMS

        no_vm_scan_since (Optional[str]): Only show assets that have not been scanned since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        no_compliance_scan_since (Optional[str]): Only show assets that have not had a compliance scan since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_since (Optional[str]): Only show assets that have been scanned since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        compliance_scan_since (Optional[str]): Only show assets that have had a compliance scan since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_processed_before (Optional[str]): Only show assets that have been processed by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_processed_after (Optional[str]): Only show assets that have been processed by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_date_before (Optional[str]): Only show assets that have been scanned by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_date_after (Optional[str]): Only show assets that have been scanned by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_auth_scan_date_before (Optional[str]): Only show assets that have been authenticated scanned by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_auth_scan_date_after (Optional[str]): Only show assets that have been authenticated scanned by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        scap_scan_since (Optional[str]): Only show assets that have been scanned by the SCAP module since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        no_scap_scan_since (Optional[str]): Only show assets that have not been scanned by the SCAP module since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.

        TAG PARAMS

        use_tags (bool): Whether to use tags. Default is 'False'. ends up being passed to API as 0 or 1.
        tag_set_by (Optional[Literal["id","name"]]): search for tags by ID or name. Default is 'id'.
        tag_include_selector (Optional[Literal["any","all"]]): Specified tags will be included if any or all are matched. Default is 'any'. Only valid if use_tags is set to 1 (True).
        tag_exclude_selector (Optional[Literal["any","all"]]): Specified tags will be excluded if any or all are matched. Default is 'any'. Only valid if use_tags is set to 1 (True).
        tag_set_include (Optional[str]): Only show assets with tags that match this string. Multiple values specified as a comma-separated string. Only valid if use_tags is set to 1 (True).
        tag_set_exclude (Optional[str]): Only show assets with tags that do not match this string. Multiple values specified as a comma-separated string. Only valid if use_tags is set to 1 (True).
        show_tags (Optional[bool]): Whether to show tags. Default is 'False'. ends up being passed to API as 0 or 1.

        ARS (Asset Risk Score) PARAMS

        show_ars (Optional[bool]): Whether to show the ARS. Default is 'False'. ends up being passed to API as 0 or 1.
        ars_min (Optional[int]): Only show assets with an ARS greater than or equal to this value.
        ars_max (Optional[int]): Only show assets with an ARS less than or equal to this value.
        show_ars_factors (Optional[bool]): Whether to show the ARS factors. Default is 'False'. ends up being passed to API as 0 or 1.

        TRURISK PARAMS

        show_trurisk (Optional[bool]): Whether to show the TruRisk. Default is 'False'. ends up being passed to API as 0 or 1.
        trurisk_min (Optional[int]): Only show assets with a TruRisk greater than or equal to this value.
        trurisk_max (Optional[int]): Only show assets with a TruRisk less than or equal to this value.
        show_trurisk_factors (Optional[bool]): Whether to show the TruRisk factors. Default is 'False'. ends up being passed to API as 0 or 1.

        CLOUD HOST PARAMS

        host_metadata (Optional[Literal["all", "ec2", "azure", "google"]]): Specify all to list all cloud and non-cloud assets. Specify a service to just get cloud assets for that service plus non-cloud assets.
        host_metadata_fields (Optional[str]): Specify the fields to return for cloud assets. Multiple values specified as a comma-separated string. Only valid if host_metadata is set to a cloud service.
        show_cloud_tags (Optional[bool]): Whether to show cloud tags. Default is 'False'. ends up being passed to API as 0 or 1.
        cloud_tag_fields (Optional[str]): Specify the cloud tag name and value, separated by a colon. Multiple values specified as a comma-separated string. If this is not specified and show_cloud_tags is 1, all tags are returned.

    Returns:

        list[Union[VMDRHost, VMDRID]]: A list of VMDRHost or VMDRID objects.
    """

    responses = BaseList()
    pulled = 0

    # add the action to the kwargs:
    kwargs["action"] = "list"

    if kwargs.get("truncation_limit") and (
        kwargs["truncation_limit"] in [0, "0"]
        and kwargs["details"] not in ["None", None]
    ):
        with LOCK:
            print(
                "[!] Warning: You have specified to pull all data with no pagination. This is generally not recommended, as it uses lots of resources and take a long time to complete. Please consider specifying a page_count or truncation_limit to avoid this issue."
            )

    while True:
        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_host_list",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )
        if response.status_code != 200:
            with LOCK:
                print("No data returned.")
            return responses

        xml = xml_parser(response.content)

        if (
            "HOST_LIST" not in xml["HOST_LIST_OUTPUT"]["RESPONSE"]
            and "ID_SET" not in xml["HOST_LIST_OUTPUT"]["RESPONSE"]
        ):
            with LOCK:
                print("No host list returned.")
            return

        # If details is none, ID_SET will be returned instead of HOST_LIST
        if "ID_SET" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]:
            # check if ID_SET is a list of dicts or a single dict:
            if isinstance(xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"], dict):
                # if it's a single dict, convert it to a list of dicts:
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"] = [
                    xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"]
                ]

            for ID in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"]:
                # create a VMDRID object and append to responses
                # This code will only run if details=None, so now we just need to check if show_asset_id is set to 1:
                if kwargs.get("show_asset_id"):
                    responses.append(VMDRID(ID=ID, TYPE="asset"))
                else:  # elif not kwargs.get("show_asset_id"):
                    responses.append(VMDRID(ID=ID, TYPE="host"))
        else:
            # HOST_LIST will be returned
            # first, check if xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]
            # is a list of dicts or a single dict:
            if isinstance(
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"], dict
            ):
                # if it's a single dict, convert it to a list of dicts:
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"] = [
                    xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]
                ]

            for host in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]:
                # return a list of VMDRHost objects
                responses.append(VMDRHost.from_dict(host))

        pulled += 1

        if page_count != "all" and pulled >= page_count:
            with LOCK:
                print(f"{current_thread.name()} Page count reached.")
            break

        if "WARNING" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]:
            if "URL" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["WARNING"]:
                with LOCK:
                    print(
                        f"{current_thread().name} (get_host_list) Pagination detected. Pulling next page from url: {xml['HOST_LIST_OUTPUT']['RESPONSE']['WARNING']['URL']}"
                    )
                # get the id_min parameter from the URL to pass into kwargs:
                params = parse_qs(
                    urlparse(
                        xml["HOST_LIST_OUTPUT"]["RESPONSE"]["WARNING"]["URL"]
                    ).query
                )
                kwargs["id_min"] = params["id_min"][0]

            else:
                break
        else:
            break

    return responses



"""
qvs.py - Contains the KBQVS class for ```get_kb_qvs``` function in the VMDR module.
"""

from typing import Union, Optional
from dataclasses import dataclass, field, asdict
from datetime import datetime

from ...base import BaseList


@dataclass(order=True)
class KBQVS:
    """
    Knowledgebase Qualys Vulnerability Severity (QVS) entry dataclass
    """

    base: Union[dict, None] = field(default=None, compare=False)  # The base entry
    # The following fields are nested under base:
    id: str = field(default=None, compare=False)
    idType: str = field(default=None, compare=False)
    qvs: Union[str, int] = field(default=None)
    qvsLastChangedDate: Union[str, datetime] = field(default=None, compare=False)
    nvdPublishedDate: Union[str, datetime] = field(default=None, compare=False)
    contributingFactors: Union[dict, None] = field(default=None, compare=False)
    # The following fields are nested under contributingFactors:
    cvss: Union[str, float] = field(default=None, init=False, compare=False)
    cvssVersion: str = field(default=None, init=False, compare=False)
    cvssString: str = field(default=None, init=False, compare=False)
    epss: Union[list, float] = field(default=None, init=False, compare=False)
    threatActors: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )  # List of one long string
    exploitMaturity: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )  # List of one long string
    trending: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )  # List of one long string
    mitigationControls: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )
    malwareName: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )  # List of one long string
    malwareHash: Optional[Union[list[str], BaseList[str]]] = field(
        default=None, compare=False
    )
    rti: Optional[Union[list[str], BaseList[str]]] = field(default=None, compare=False)

    def __post_init__(self):
        if not self.base:
            raise ValueError("The base attribute is required.")

        DT_FIELDS = ["qvsLastChangedDate", "nvdPublishedDate"]
        for field in DT_FIELDS:
            if field and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromtimestamp(int(self.base.get(field))))

        if self.base.get("id"):
            setattr(self, "id", self.base.get("id"))

        if self.base.get("qvs") and not isinstance(self.base.get("qvs"), int):
            setattr(self, "qvs", int(self.base.get("qvs")))

        if self.base.get("idType"):
            setattr(self, "idType", self.base.get("idType"))

        if self.contributingFactors and isinstance(self.contributingFactors, dict):
            if "epss" in self.contributingFactors.keys():
                setattr(self, "epss", float(self.contributingFactors.get("epss")[0]))

            if "cvssVersion" in self.contributingFactors.keys():
                setattr(
                    self, "cvssVersion", self.contributingFactors.get("cvssVersion")
                )

            if "cvssString" in self.contributingFactors.keys():
                setattr(self, "cvssString", self.contributingFactors.get("cvssString"))

            if "cvss" in self.contributingFactors.keys():
                setattr(self, "cvss", float(self.contributingFactors.get("cvss")))

            ONE_LONG_STR_FIELDS = [
                "threatActors",
                "exploitMaturity",
                "trending",
                "malwareName",
            ]
            for field in ONE_LONG_STR_FIELDS:
                if field in self.contributingFactors.keys():
                    bl = BaseList()
                    for item in self.contributingFactors.get(field)[0].split(","):
                        if field == "trending":
                            item = int(item)
                        bl.append(item)
                    setattr(self, field, bl)

            if "mitigationControls" in self.contributingFactors.keys():
                bl = BaseList()
                for control in self.contributingFactors.get("mitigationControls")[0][
                    0
                ].split(","):
                    bl.append(control)
                setattr(self, "mitigationControls", bl)

            if "malwareHash" in self.contributingFactors.keys():
                bl = BaseList()
                for h in self.contributingFactors.get("malwareHash"):
                    bl.append(h)
                setattr(self, "malwareHash", bl)

            if "rti" in self.contributingFactors.keys():
                bl = BaseList()
                for r in self.contributingFactors.get("rti"):
                    bl.append(r)
                setattr(self, "rti", bl)

        # Set key fields to None
        setattr(self, "contributingFactors", None)
        setattr(self, "base", None)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def __str__(self):
        return str(self.id)

    def __int__(self):
        return self.qvs

    def from_dict(data: dict):
        return KBQVS(**data)

    def epss_category(self):
        """
        Returns a low/medium/high/critical category based on the EPSS value.

        Returns:
            str: The category of the EPSS value.
        """
        if self.epss < 0.15:
            return "Low"
        elif self.epss < 0.4:
            return "Medium"
        elif self.epss < 0.7:
            return "High"
        elif self.epss < 1:
            return "Critical"
        else:
            return "Unknown"

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()



"""
Contains the Container dataclass.
"""

from dataclasses import dataclass, asdict
from typing import Union, Literal
from datetime import datetime
from ipaddress import ip_address

from ...base.base_list import BaseList
from ...exceptions.Exceptions import *


@dataclass
class Container:
    """
    Represents a Docker container.
    """

    imageId: str = None
    imageUuid: str = None
    containerId: str = None
    created: Union[str, datetime] = None
    updated: Union[str, datetime] = None
    label: str = None
    uuid: str = None
    sha: str = None
    privileged: bool = None
    path: str = None
    imageSha: str = None
    macAddress: str = None
    customerUuid: str = None
    ipv4: ip_address = None
    ipv6: ip_address = None
    name: str = None
    host: dict = None
    # host is parsed into below fields:
    host_sensorUuid: str = None
    host_hostname: str = None
    host_ipAddress: ip_address = None
    host_uuid: str = None
    host_lastUpdated: Union[str, datetime] = None
    # End host fields
    hostArchitecture: str = None
    state: str = None
    portMapping: list[dict] = None
    stateChanged: Union[str, datetime] = None
    services: list[dict] = None
    operatingSystem: str = None
    lastScanned: Union[str, datetime] = None
    source: str = None
    environment: str = None
    arguments: str = None
    command: str = None
    drift: dict = None
    vulnerabilities: dict = None
    softwares: dict = None
    isDrift: bool = None
    isRoot: bool = None
    cluster: str = None
    users: list[dict] = None

    def __post_init__(self):
        """
        Post-initialization function for the Container class.
        """

        # Convert any datetime strings to datetime objects using fromtimestamp:
        DT_FIELDS = ["created", "updated", "stateChanged", "lastScanned"]
        for field in DT_FIELDS:
            if isinstance(getattr(self, field), str):
                setattr(
                    self,
                    field,
                    datetime.fromtimestamp(int(getattr(self, field)) / 1000),
                )

        # Convert any IP address strings to ip_address objects:
        IP_FIELDS = ["ipv4", "ipv6"]
        for field in IP_FIELDS:
            if isinstance(getattr(self, field), str):
                setattr(self, field, ip_address(getattr(self, field)))

        if self.label:
            pass

        if self.path:
            pass

        if self.macAddress:
            pass

        if self.ipv4:
            pass

        if self.ipv6:
            pass

        if self.host:
            host_dt_fields = ["lastUpdated"]
            for field in host_dt_fields:
                if isinstance(self.host.get(field), str):
                    setattr(
                        self.host,
                        f"host_{field}",
                        datetime.fromtimestamp(int(self.host[field]) / 1000),
                    )

            if self.host.get("ipAddress"):
                setattr(self, "host_ipAddress", ip_address(self.host["ipAddress"]))

            host_fields = ["sensorUuid", "hostname", "uuid"]
            for field in host_fields:
                if self.host.get(field):
                    setattr(self, f"host_{field}", self.host[field])

            # Set the original host field to None:
            setattr(self, "host", None)

        # NOTE: I have yet to see any of the commented
        # out fields below. I will update this as needed.
        """
        if self.hostArchitecture:
            pass

        if self.services:
            pass

        if self.users:
            pass

        if self.operatingSystem:
            pass

        if self.lastScanned:
            pass

        if self.environment:
            pass

        if self.arguments:
            pass

        if self.drift:
            pass

        if self.vulnerabilities:
            pass

        if self.softwares:
            pass

        if self.isRoot:
            pass

        if self.cluster:
            pass
        """

        if self.portMapping:
            data = self.portMapping
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for item in data:
                bl.append(item)
            setattr(self, "portMapping", bl)

    def has_drift(self) -> bool:
        """
        Check if the container has drift.

        Returns:
            bool: True if the container has drift, False otherwise.
        """
        return self.isDrift

    def is_root(self) -> bool:
        """
        Check if the container is running as root.

        Returns:
            bool: True if the container is running as root, False otherwise.
        """
        return self.isRoot

    def __dict__(self) -> dict:
        """
        Return the object as a dictionary.

        Returns:
            dict: The object as a dictionary.
        """
        return asdict(self)

    def to_dict(self) -> dict:
        """
        Return the object as a dictionary.

        Returns:
            dict: The object as a dictionary.
        """
        return asdict(self)

    def keys(self) -> list[str]:
        """
        Return the keys of the object.

        Returns:
            list[str]: The keys of the object.
        """
        return self.to_dict().keys()

    def values(self) -> list:
        """
        Return the values of the object.

        Returns:
            list: The values of the object.
        """
        return self.to_dict().values()

    def items(self) -> list[tuple]:
        """
        Return the items of the object.

        Returns:
            list[tuple]: The items of the object.
        """
        return self.to_dict().items()



"""
Interact with connectors for a given cloud provider.
"""

from typing import Union, Literal

from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.Connectors import AWSConnector, AzureConnector


def get_connectors(
    auth: BasicAuth,
    provider: Literal["aws", "azure"],
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> BaseList[Union[AWSConnector, AzureConnector]]:
    """
    Get all Connector definitions from the Qualys CloudView API

    Params:
        auth (BasicAuth): The authentication object.
        provider (Literal['aws', 'azure']): The cloud provider to get connectors for.
        page_count (int): The number of pages to return. If 'all', return all pages. Default is 'all'.

    ## Kwargs:

        pageNo (int): The ordered page to start retrieving connectors from, or if page_count is 1, the page to retrieve.
        pageSize (int): The number of connectors to get per page.
        filter (str): Filter connectors by providing a query. Queryable fields are: name, description, state (SUCCESS, PENDING, REGIONS_DISCOVERED, ERROR), connector.uuid, lastSyncedOn (in UTC). Example: filter="name:MyConnector"
        sort (Literal['lastSyncedOn:asc', 'lastSyncedOn:desc']): Sort the connectors by lastSyncedOn in ascending or descending order.

    Returns:
        BaseList[Connector]: The response from the API as a BaseList of <provider>Connector objects.
    """

    # Check cloud provider is valid
    provider = provider.lower()
    if provider not in ["aws", "azure"]:
        raise QualysAPIError(
            f"Invalid provider {provider}. Valid providers: 'aws' or 'azure'"
        )

    responses = BaseList()
    currentPage = 0

    if kwargs.get("filter"):
        # Make sure the query key is valid
        if not kwargs["filter"].split(":")[0] in [
            "name",
            "description",
            "state",
            "connector.uuid",
            "lastSyncedOn",
        ]:
            raise QualysAPIError(
                "Invalid filter key. Valid keys: name, description, state, connector.uuid, lastSyncedOn"
            )
        # If using state, make sure value is UPPERCASE:
        if kwargs["filter"].split(":")[0] == "state":
            kwargs[
                "filter"
            ] = f"{kwargs['filter'].split(':')[0]}:{kwargs['filter'].split(':')[1].upper()}"

    while True:
        # Set the current page number and page size in kwargs
        kwargs["pageNo"] = currentPage
        # Set the cloudprovider in kwargs so it is appended to the URL path
        kwargs["cloudprovider"] = provider

        # Make the API request to retrieve the connectors
        response = call_api(
            auth=auth, module="cloudview", endpoint=f"get_connectors", params=kwargs
        )

        if response.status_code != 200:
            raise QualysAPIError(
                f"Error retrieving {provider} connectors. Status code: {response.status_code}. Requests reporting {response.reason}"
            )

        j = response.json()

        if len(j["content"]) == 0:
            print("No connectors found.")
            break

        # Iterate through the records in the response and create Connector objects
        for record in j["content"]:
            match provider:
                case "aws":
                    responses.append(AWSConnector(**record))
                case "azure":
                    responses.append(AzureConnector(**record))

        # Print a message indicating the current page was retrieved successfully
        print(f"Page {currentPage+1} of {provider} connectors retrieved successfully.")
        currentPage += 1

        # Break the loop if all pages are retrieved or the requested number of pages are retrieved
        if (page_count != "all" and currentPage + 1 > page_count) or j["last"]:
            break

    # Print a message indicating all pages have been retrieved
    print(
        f"All pages complete. {str(len(responses))} {provider} connector records retrieved."
    )

    return responses


def get_connector_details(
    auth: BasicAuth, provider: Literal["aws", "azure"], connectorId: str
) -> Union[AWSConnector, AzureConnector]:
    """
    Get details for a single connector by connectorId

    Params:
        auth (BasicAuth): The authentication object.
        provider (Literal['aws', 'azure']): The cloud provider to get connectors for.
        connectorId (str): The connectorId of the connector to get details for.

    Returns:
        Connector: The response from the API as a <provider>Connector object.
    """

    # Check cloud provider is valid
    provider = provider.lower()
    if provider not in ["aws", "azure"]:
        raise QualysAPIError(
            f"Invalid provider {provider}. Valid providers: 'aws' or 'azure'"
        )

    response = call_api(
        auth=auth,
        module="cloudview",
        endpoint=f"get_connector_details",
        params={
            "placeholder": connectorId,
            "cloudprovider": provider,
        },  # placeholders let us append the connectorId/cloud provider to the URL path
    )

    if response.status_code != 200:
        raise QualysAPIError(
            f"Error retrieving {provider} connector details. Status code: {response.status_code}"
        )

    j = response.json()

    match provider:
        case "aws":
            return AWSConnector(**j)
        case "azure":
            return AzureConnector(**j)
        case _:
            raise QualysAPIError(
                f"Invalid provider {provider}. Valid providers: 'aws' or 'azure'"
            )


# BEGIN PLATFORM SPECIFIC FUNCTIONS:

"""AWS FUNCTIONS"""


def get_aws_base_account(auth: BasicAuth) -> dict:
    """
    Get the AWS Base Account details

    Params:
        auth (BasicAuth): The authentication object.

    Returns:
        dict: The response from the API as a dictionary.
    """

    response = call_api(
        auth=auth,
        module="cloudview",
        endpoint="get_aws_base_account",
    )

    if response.status_code != 200:
        raise QualysAPIError(
            f"Error retrieving AWS base account details. Status code: {response.status_code}"
        )

    return response.json()



"""
Contains resource dataclasses for AWS, such as buckets, EC2s, etc.
"""

from dataclasses import dataclass, asdict
from typing import Union
from datetime import datetime
from json import loads
from urllib.parse import unquote

from ...base.base_list import BaseList
from .QID import QID


@dataclass
class BaseResource:
    """
    Base dataclass for all AWS resources in TotalCloud
    """

    resourceId: str = None
    name: str = None
    connectorUuids: BaseList[str] = None
    created: Union[str, datetime] = None
    cloudAccountId: int = None
    additionalDetails: dict = None
    uuid: str = None
    connectorUuid: str = None
    createdOn: Union[str, datetime] = None
    tags: BaseList[str] = None
    remediationEnabled: bool = None
    lastUpdated: Union[str, datetime] = None
    cloudType: str = None
    region: str = None
    resourceType: str = None
    accountAlias: str = None
    controlsFailed: int = None
    qualysTags: BaseList[str] = None

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the BaseResource dataclass
        """
        DT_FIELDS = ["created", "createdOn", "lastUpdated"]
        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.additionalDetails:
            setattr(self, "additionalDetails", str(self.additionalDetails))

        if self.connectorUuids:
            data = self.connectorUuids
            bl = BaseList()
            for uuid in data:
                bl.append(uuid)
            setattr(self, "connectorUuids", bl)

        if self.tags:
            data = self.tags
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(f"{tag.get('key', None)}:{tag.get('value', None)}")
            setattr(self, "tags", bl)

        if self.qualysTags:
            data = self.qualysTags
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(tag["tagName"])
            setattr(self, "qualysTags", bl)

        if self.cloudAccountId:
            setattr(self, "cloudAccountId", int(self.cloudAccountId))

        if self.remediationEnabled:
            data = self.remediationEnabled
            if not isinstance(data, bool):
                setattr(self, "remediationEnabled", data.lower() == "true")

    def to_dict(self):
        """
        to_dict - return the BaseResource as a dictionary
        """
        return asdict(self)

    def keys(self):
        """
        keys - return the keys of the BaseResource
        """
        return self.to_dict().keys()

    def values(self):
        """
        values - return the values of the BaseResource
        """
        return self.to_dict().values()

    def items(self):
        """
        items - return the items of the BaseResource
        """
        return self.to_dict().items()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create a BaseResource object from a dictionary
        """
        return BaseResource(**data)


@dataclass
class AWSBucket(BaseResource):
    """
    Represents an S3 bucket in AWS
    """

    bucketName: str = None
    bucketCreationDateStr: Union[str, datetime] = None
    s3GrantList: BaseList[str] = None
    ownerName: str = None
    bucketPolicy: str = None
    bucketOwnerId: str = None

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the AWSBucket dataclass
        """

        DT_FIELDS = ["bucketCreationDateStr"]
        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.s3GrantList:
            data = self.s3GrantList
            bl = BaseList()
            for grant in data:
                bl.append(
                    f"(email: {grant.get('emailAddress', None)}, groupUri: {grant.get('groupUri', None)}, displayName: {grant.get('displayName', None)}, permission: {grant.get('permission', None)}, id: {grant.get('id', None)})"
                )
            setattr(self, "s3GrantList", bl)

        # For some reason, Qualys returns bool fields as strings for certain resources?
        BOOL_FIELDS = ["remediationEnabled"]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSBucket object from a dictionary
        """
        return AWSBucket(**data)


@dataclass
class AWSNetworkACL(BaseResource):
    """
    Represents a Network ACL in AWS
    """

    associations: BaseList[str] = None
    ipPermissionEgressList: BaseList[str] = None
    ipPermissionList: BaseList[str] = None
    isDefault: bool = None
    vpcId: str = None
    networkAclId: str = None

    def __post_init__(self):
        if self.associations:
            data = self.associations
            bl = BaseList()
            for association in data:
                bl.append(
                    f"(subnetId: {association.get('subnetId', None)}, networkAclAssociationId: {association.get('networkAclAssociationId', None)}, networkAclId: {association.get('networkAclId', None)})"
                )
            setattr(self, "associations", bl)

        if self.ipPermissionEgressList:
            data = self.ipPermissionEgressList
            bl = BaseList()
            for permission in data:
                portRange = permission.get("portRange", None)
                ruleAction = permission.get("ruleAction", None)
                protocol = permission.get("protocol", None)
                source = permission.get("source", None)
                _type = permission.get("type", None)
                s = f"(portRange: {portRange}, ruleAction: {ruleAction}, protocol: {protocol}, source: {source}, type: {_type})"
                bl.append(s)
            setattr(self, "ipPermissionEgressList", bl)

        if self.ipPermissionList:
            data = self.ipPermissionList
            bl = BaseList()
            for permission in data:
                portRange = permission.get("portRange", None)
                ruleAction = permission.get("ruleAction", None)
                protocol = permission.get("protocol", None)
                source = permission.get("source", None)
                _type = permission.get("type", None)
                s = f"(portRange: {portRange}, ruleAction: {ruleAction}, protocol: {protocol}, source: {source}, type: {_type})"
                bl.append(s)
            setattr(self, "ipPermissionList", bl)

        BOOL_FIELDS = ["isDefault"]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSNetworkACL object from a dictionary
        """
        return AWSNetworkACL(**data)


@dataclass
class AWSRDS(BaseResource):
    """
    Represents a Relational Database Service in AWS
    """

    subnetId: BaseList[str] = None
    subnetGroup: None = None  # Gets parsed out
    # subnetGroup is parsed out to below fields
    subnetGroup_dbSubnetGroupStatus: str = None
    subnetGroup_dbSubnetGroupName: str = None
    subnetGroup_dbSubnetVpcId: str = None
    subnetGroup_subnetList: BaseList[str] = None
    subnetGroup_dbsubnetGroupArn: str = None
    # end of subnetGroup fields
    dbName: str = None
    availabilityZone: str = None
    securityGroupId: BaseList[str] = None
    dbClusterIdentifier: str = None
    caCertificateIdentifier: str = None
    engineVersion: str = None
    masterUsername: str = None
    dbInstanceIdentifier: str = None
    dbSecurityGroupList: BaseList[str] = None
    # name: str = None
    backupRetentionPeriod: int = None
    kmsKeyId: str = None
    status: str = None
    latestRestorableTime: Union[str, datetime] = None
    dbInstancePort: int = None
    enhancedMonitoringEnabled: bool = None
    dbiResourceId: str = None
    preferredBackupWindow: str = None
    endpoint: None = None  # Gets parsed out
    # endpoint is parsed out to below fields
    endpoint_hostedZoneId: str = None
    endpoint_address: str = None
    endpoint_port: int = None
    # end of endpoint fields
    engine: str = None
    publiclyAccessible: bool = None
    arn: str = None
    instanceCreatedTime: Union[str, datetime] = None
    multiAZ: bool = None
    instanceClass: str = None
    encrypted: bool = None
    iamDatabaseAuthenticationEnabled: bool = None
    licenseModel: str = None
    preferredMaintenanceWindow: str = None
    storageType: str = None

    def __post_init__(self):
        if self.subnetId:
            data = self.subnetId
            bl = BaseList()
            for subnet in data:
                bl.append(subnet)
            setattr(self, "subnetId", bl)

        if self.subnetGroup:
            data = self.subnetGroup
            for key, value in data.items():
                if key == "subnetList":
                    bl = BaseList()
                    for subnet in value:
                        bl.append(
                            f"(identifier: {subnet.get('identifier', None)}, availabilityZone: {subnet.get('availabilityZone', None)}, status: {subnet.get('status', None)})"
                        )
                    setattr(self, "subnetGroup_subnetList", bl)
                else:
                    setattr(self, f"subnetGroup_{key}", value)
            setattr(self, "subnetGroup", None)

        if self.securityGroupId:
            data = self.securityGroupId
            bl = BaseList()
            for group in data:
                bl.append(group)
            setattr(self, "securityGroupId", bl)

        if self.dbSecurityGroupList:
            data = self.dbSecurityGroupList
            bl = BaseList()
            for group in data:
                bl.append(group)
            setattr(self, "dbSecurityGroupList", bl)

        if self.endpoint:
            data = self.endpoint
            for key, value in data.items():
                setattr(self, f"endpoint_{key}", value)
            setattr(self, "endpoint", None)

        DT_FIELDS = ["latestRestorableTime", "instanceCreatedTime"]
        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        BOOL_FIELDS = [
            "enhancedMonitoringEnabled",
            "multiAZ",
            "encrypted",
            "iamDatabaseAuthenticationEnabled",
            "publiclyAccessible",
        ]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSRDS object from a dictionary
        """
        return AWSRDS(**data)


@dataclass
class AWSIAMUser(BaseResource):
    """
    Represents an IAM User in AWS
    """

    userPolicies: BaseList[str] = None
    userAttachedPolicies: BaseList[str] = None
    userDto: None = None  # Gets parsed out
    # userDto is parsed out to below fields
    userDto_accessKey2Active: bool = None
    userDto_cert2LastRotated: Union[str, datetime] = None
    userDto_cert2Active: bool = None
    userDto_accessKey2LastRotated: Union[str, datetime] = None
    userDto_path: str = None
    userDto_accessKeys: BaseList[str] = None
    userDto_passwordLastUsed: Union[str, datetime] = None
    userDto_passwordNextRotation: Union[str, datetime] = None
    userDto_accessKey1LastUsedRegion: str = None
    userDto_arn: str = None
    userDto_userCreationTime: Union[str, datetime] = None
    userDto_accessKey1LastRotated: Union[str, datetime] = None
    userDto_cert1Active: bool = None
    userDto_mfaActive: bool = None
    userDto_accessKey2LastUsedService: str = None
    userDto_passwordEnabled: bool = None
    userDto_accessKey1Active: bool = None
    userDto_accessKey1LastUsed: Union[str, datetime] = None
    userDto_accessKey2LastUsed: Union[str, datetime] = None
    userDto_userId: str = None
    userDto_createdDate: Union[str, datetime] = None
    userDto_passwordLastChanged: Union[str, datetime] = None
    userDto_cert1LastRotated: Union[str, datetime] = None
    userDto_accessKey1LastUsedService: str = None
    userDto_accessKey2LastUsedRegion: str = None
    userDto_username: str = None
    # end of userDto fields
    userGroups: BaseList[str] = None
    classifications: BaseList[str] = None
    userInlinePolicies: BaseList[str] = None
    user: str = None

    def __post_init__(self):
        POLICY_FIELDS = ["userPolicies", "userAttachedPolicies"]
        for field in POLICY_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for policy in data:
                    if field == "userAttachedPolicies":
                        bl.append(
                            f"(policyArn: {policy.get('policyArn', None)}, policyName: {policy.get('policyName', None)})"
                        )
                    else:
                        bl.append(policy)

                setattr(self, field, bl)

        if self.userDto:
            data = self.userDto
            bl = BaseList()
            # First, parse out accessKeys:
            accessKeys = data.get("accessKeys", None)
            accessKeysbl = BaseList()
            for key in accessKeys:
                timestamp = key.get("createDate", None)
                if timestamp:  # format: 1703001007000
                    timestamp = datetime.fromtimestamp(int(timestamp) / 1000)
                status = key.get("status", None)
                accessKeyId = key.get("accessKeyId", None)
                accessKeysbl.append(
                    f"(accessKeyId: {accessKeyId}, createDate: {timestamp}, status: {status})"
                )
            setattr(self, "userDto_accessKeys", accessKeysbl)
            # Now we can parse the rest of the fields
            for key, value in data.items():
                if key == "accessKeys":
                    continue
                else:
                    setattr(self, f"userDto_{key}", value)
            setattr(self, "userDto", None)

        if self.userGroups:
            data = self.userGroups
            bl = BaseList()
            for group in data:
                groupArn = group.get("groupArn", None)
                path = group.get("path", None)
                groupName = group.get("groupName", None)
                createdDate = group.get("createdDate", None)
                if createdDate:
                    createdDate = datetime.fromtimestamp(int(createdDate) / 1000)
                groupId = group.get("groupId", None)
                bl.append(
                    f"(groupArn: {groupArn}, path: {path}, groupName: {groupName}, createdDate: {createdDate}, groupId: {groupId})"
                )
            setattr(self, "userGroups", bl)

        if self.classifications:
            data = self.classifications
            bl = BaseList()
            for classification in data:
                bl.append(classification)
            setattr(self, "classifications", bl)

        if self.userInlinePolicies:
            data = self.userInlinePolicies
            bl = BaseList()
            for policy in data:
                policyDocument = policy.get("policyDocument", None)
                policyName = policy.get("policyName", None)
                bl.append(
                    f"(policyDocument: {policyDocument}, policyName: {policyName})"
                )
            setattr(self, "userInlinePolicies", bl)

        DT_FIELDS = [
            "userDto_cert2LastRotated",
            "userDto_accessKey2LastRotated",
            "userDto_passwordLastUsed",
            "userDto_passwordNextRotation",
            "userDto_userCreationTime",
            "userDto_accessKey1LastRotated",
            "userDto_createdDate",
            "userDto_passwordLastChanged",
            "userDto_cert1LastRotated",
            "userDto_accessKey1LastUsed",
            "userDto_accessKey2LastUsed",
        ]
        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(
                    self,
                    field,
                    datetime.fromtimestamp(int(getattr(self, field)) / 1000),
                )

        BOOL_FIELDS = [
            "userDto_accessKey2Active",
            "userDto_cert2Active",
            "userDto_cert1Active",
            "userDto_mfaActive",
            "userDto_passwordEnabled",
            "userDto_accessKey1Active",
        ]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSIAMUser object from a dictionary
        """
        return AWSIAMUser(**data)


@dataclass
class AWSVPC(BaseResource):
    """
    Represents a Virtual Private Cloud in AWS
    """

    instanceTenancy: str = None
    cidrBlock: str = None
    isDefault: bool = None
    vpcId: str = None
    ipv6CidrBlockAssociationSet: BaseList[str] = None

    def __post_init__(self):
        if self.ipv6CidrBlockAssociationSet:
            data = self.ipv6CidrBlockAssociationSet
            bl = BaseList()
            for association in data:
                associationId = association.get("associationId", None)
                ipv6CidrBlockState = association.get("ipv6CidrBlockState", None)
                ipv6CidrBlock = association.get("ipv6CidrBlock", None)
                bl.append(
                    f"(associationId: {associationId}, ipv6CidrBlockState: {ipv6CidrBlockState}, ipv6CidrBlock: {ipv6CidrBlock})"
                )
            setattr(self, "ipv6CidrBlockAssociationSet", bl)

        BOOL_FIELDS = ["isDefault"]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSVPC object from a dictionary
        """
        return AWSVPC(**data)


@dataclass
class AWSSecurityGroup(BaseResource):
    """
    Represents a security group in AWS
    """

    groupId: str = None
    description: str = None
    vulnerable: bool = None
    vpcId: str = None
    ipPermissionList: BaseList[str] = None
    ipPermissionEgressList: BaseList[str] = None
    groupName: str = None

    def __post_init__(self):
        if self.ipPermissionList:
            data = self.ipPermissionList
            bl = BaseList()
            for permission in data:
                portRange = permission.get("portRange", None)
                ruleAction = permission.get("ruleAction", None)
                protocol = permission.get("protocol", None)
                source = permission.get("source", None)
                _type = permission.get("type", None)
                s = f"(portRange: {portRange}, ruleAction: {ruleAction}, protocol: {protocol}, source: {source}, type: {_type})"
                bl.append(s)
            setattr(self, "ipPermissionList", bl)

        if self.ipPermissionEgressList:
            data = self.ipPermissionEgressList
            bl = BaseList()
            for permission in data:
                portRange = permission.get("portRange", None)
                ruleAction = permission.get("ruleAction", None)
                protocol = permission.get("protocol", None)
                source = permission.get("source", None)
                _type = permission.get("type", None)
                s = f"(portRange: {portRange}, ruleAction: {ruleAction}, protocol: {protocol}, source: {source}, type: {_type})"
                bl.append(s)
            setattr(self, "ipPermissionEgressList", bl)

        # For some reason, Qualys returns bool fields as strings for certain resources?
        BOOL_FIELDS = ["vulnerable"]
        for field in BOOL_FIELDS:
            data = getattr(self, field)
            if data and not isinstance(data, bool):
                setattr(self, field, data.lower() == "true")

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSSecurityGroup object from a dictionary
        """
        return AWSSecurityGroup(**data)


@dataclass
class AWSLambdaFunction(BaseResource):
    """
    Represents a Lambda function in AWS
    """

    associations: BaseList[str] = None
    handler: str = None
    rawFunctionPolicy: str = None
    masterArn: str = None
    aliases: BaseList[str] = None
    assumeRolePolicyDocument: str = None
    kmsKeyArn: str = None
    tracingConfig: str = None
    description: str = None
    inlinePolicies: BaseList[str] = None
    codeSize: int = None
    timeout: int = None
    codeSha256: str = None
    roleArn: str = None
    vpcId: str = None
    layers: BaseList[str] = None
    functionArn: str = None
    functionVersions: BaseList[str] = None
    functionName: str = None
    runtime: str = None
    functionUrlConfigs: BaseList[str] = None
    revisionId: str = None
    classifications: BaseList[str] = None
    managedPolicies: BaseList[str] = None
    memorySize: int = None
    eventSourceMappings: BaseList[str] = None
    roleName: str = None
    lastModified: Union[str, datetime] = None
    version: str = None

    def __post_init__(self):
        # all fields above with a comment should be investigated!

        if self.aliases:
            data = self.aliases
            bl = BaseList()
            for alias in data:
                revisionId = alias.get("revisionId", None)
                aliasArn = alias.get("aliasArn", None)
                routingConfigAdditionalVersion = alias.get(
                    "routingConfigAdditionalVersion", None
                )
                functionVersion = alias.get("functionVersion", None)
                name = alias.get("name", None)
                description = alias.get("description", None)
                routingConfigAdditionalVersionWeight = alias.get(
                    "routingConfigAdditionalVersionWeight", None
                )
                bl.append(
                    f"(revisionId: {revisionId}, aliasArn: {aliasArn}, routingConfigAdditionalVersion: {routingConfigAdditionalVersion}, functionVersion: {functionVersion}, name: {name}, description: {description}, routingConfigAdditionalVersionWeight: {routingConfigAdditionalVersionWeight})"
                )
            setattr(self, "aliases", bl)

        if self.inlinePolicies:
            data = self.inlinePolicies
            bl = BaseList()
            for policy in data:
                policyDocument_version = policy["policyDocument"]
                policyDocument_policyName = policy.get("policyName", None)
                policyDocument = f"(version: {policyDocument_version}, policyName: {policyDocument_policyName})"
                bl.append(policyDocument)
            setattr(self, "inlinePolicies", bl)

        if self.layers:
            data = self.layers
            bl = BaseList()
            for layer in data:
                name = layer.get("name", None)
                version = layer.get("version", None)
                codeSize = layer.get("codeSize", None)
                bl.append(f"(name: {name}, version: {version}, codeSize: {codeSize})")
            setattr(self, "layers", bl)

        if self.functionVersions:
            data = self.functionVersions
            bl = BaseList()
            for version in data:
                role = version.get("role", None)
                memorySize = version.get("memorySize", None)
                runtime = version.get("runtime", None)
                lastModified = version.get("lastModified", None)
                _version = version.get("version", None)
                codeSize = version.get("codeSize", None)
                bl.append(
                    f"(role: {role}, memorySize: {memorySize}, runtime: {runtime}, lastModified: {lastModified}, version: {_version}, codeSize: {codeSize})"
                )
            setattr(self, "functionVersions", bl)

        if self.functionUrlConfigs:
            data = self.functionUrlConfigs
            bl = BaseList()
            for config in data:
                lastModifiedTime = config.get("lastModifiedTime", None)
                cors = config.get("cors", None)
                creationTime = config.get("creationTime", None)
                functionUrl = config.get("functionUrl", None)
                authType = config.get("authType", None)
                functionArn = config.get("functionArn", None)
                bl.append(
                    f"(lastModifiedTime: {lastModifiedTime}, cors: {cors}, creationTime: {creationTime}, functionUrl: {functionUrl}, authType: {authType}, functionArn: {functionArn})"
                )
            setattr(self, "functionUrlConfigs", bl)

        if self.classifications:
            data = self.classifications
            bl = BaseList()
            for classification in data:
                bl.append(classification)
            setattr(self, "classifications", bl)

        if self.managedPolicies:
            data = self.managedPolicies
            bl = BaseList()
            for policy in data:
                policyArn = policy.get("policyArn", None)
                policyDocument = policy.get("policyDocument", None)
                policyName = policy.get("policyName", None)
                bl.append(
                    f"(policyArn: {policyArn}, policyDocument: {policyDocument}, policyName: {policyName})"
                )
            setattr(self, "managedPolicies", bl)

        if self.eventSourceMappings:
            data = self.eventSourceMappings
            bl = BaseList()
            for mapping in data:
                eventSourceArn = mapping.get("eventSourceArn", None)
                lastProcessingResult = mapping.get("lastProcessingResult", None)
                state = mapping.get("state", None)
                lastModified = mapping.get("lastModified", None)
                _type = mapping.get("type", None)
                bl.append(
                    f"(eventSourceArn: {eventSourceArn}, lastProcessingResult: {lastProcessingResult}, state: {state}, lastModified: {lastModified}, type: {_type})"
                )
            setattr(self, "eventSourceMappings", bl)

        if self.lastModified:
            if not isinstance(self.lastModified, datetime):
                setattr(self, "lastModified", datetime.fromisoformat(self.lastModified))

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSLambdaFunction object from a dictionary
        """
        return AWSLambdaFunction(**data)


@dataclass
class AWSSubnet(BaseResource):
    """
    Represents a subnet in AWS
    """

    subnetId: str = None
    availabilityZone: str = None
    assignIpv6AddressOnCreation: bool = None
    vpcId: str = None
    ipv6CidrBlockAssociationSet: BaseList[str] = None
    cidrBlock: str = None
    availableIpAddressCount: int = None
    mapPublicIpOnLaunch: bool = None
    defaultForAz: bool = None

    def __post_init__(self):
        if self.ipv6CidrBlockAssociationSet:
            data = self.ipv6CidrBlockAssociationSet
            bl = BaseList()
            for association in data:
                if isinstance(association, dict):
                    # None of the subnets i have seen have this field,
                    # but we can take a guess?
                    # Get the keys, and add to a string
                    s = ""
                    for key in association.keys():
                        s += f"{key}: {association[key]}, "
                    # Take out the last comma and space, and wrap s in ()s
                    s = f"({s[:-2]})"
                    bl.append(s)
                else:
                    bl.append(str(association))
            setattr(self, "ipv6CidrBlockAssociationSet", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSSubnet object from a dictionary
        """
        return AWSSubnet(**data)


@dataclass
class AWSInternetGateway(BaseResource):
    """
    Represents an Internet Gateway in AWS
    """

    attachments: BaseList[str] = None
    internetGatewayId: str = None

    def __post_init__(self):
        if self.attachments:
            data = self.attachments
            bl = BaseList()
            for attachment in data:
                s = ""
                for key in attachment.keys():
                    s += f"{key}: {attachment[key]}, "
                s = f"({s[:-2]})"
                bl.append(s)
            setattr(self, "attachments", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSInternetGateway object from a dictionary
        """
        return AWSInternetGateway(**data)


@dataclass
class AWSLoadBalancer(BaseResource):
    """
    Represents a Load Balancer in AWS
    """

    loadBalancerArn: str = None
    scheme: str = None
    instances: BaseList[str] = None  # ???
    subnetAvailabilityZonePair: BaseList[str] = None
    dnsName: str = None
    availabilityZones: BaseList[str] = None
    _type: str = None
    loadBalancerName: str = None
    createdTime: Union[str, datetime] = None
    subnets: BaseList[str] = None
    state: str = None
    listeners: BaseList[str] = None
    ipAddressType: str = None
    vpcid: str = None
    securityGroups: BaseList[AWSSecurityGroup] = None  # Double check!

    def __post_init__(self):
        BL_STR_FIELDS = [
            "subnetAvailabilityZonePair",
            "availabilityZones",
            "subnets",
            "instances",
            "securityGroups",
        ]
        for field in BL_STR_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    bl.append(item)
                setattr(self, field, bl)

        if self.listeners:
            data = self.listeners
            bl = BaseList()
            for listener in data:
                s = ""
                for key in listener.keys():
                    s += f"{key}: {listener[key]}, "
                s = f"({s[:-2]})"
                bl.append(s)
            setattr(self, "listeners", bl)

        if self.createdTime:
            if not isinstance(self.createdTime, datetime):
                setattr(self, "createdTime", datetime.fromisoformat(self.createdTime))

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSLoadBalancer object from a dictionary
        """
        return AWSLoadBalancer(**data)


@dataclass
class AWSEC2Instance(BaseResource):
    """
    Represents an EC2 Instance in AWS
    """

    subnetId: str = None
    arsScore: int = None
    availabilityZone: str = None
    instanceId: str = None
    instanceState: str = None
    networkInterfaceAddresses: BaseList[str] = None  # list of dicts
    vulnerabilityStats: None = None  # parse out to below fields
    # vulnerabilityStats is parsed out to below fields
    vulnerabilityStats_severity_1: int = 0
    vulnerabilityStats_severity_2: int = 0
    vulnerabilityStats_severity_3: int = 0
    vulnerabilityStats_severity_4: int = 0
    vulnerabilityStats_severity_5: int = 0
    vulnerabilityStats_typeDetected_Confirmed: int = 0
    vulnerabilityStats_typeDetected_Potential: int = 0
    vulnerabilityStats_typeDetected_Informational: int = 0
    vulnerabilityStats_typeDetected_totalVulnerability: int = 0
    # end of vulnerabilityStats fields
    vpcId: str = None
    events: BaseList[str] = None  # ???
    iamInstanceProfileRoleDetails: None = None  # parse out to below fields
    # iamInstanceProfileRoleDetails is parsed out to below fields
    iamInstanceProfileRoleDetails_profileName: str = None
    iamInstanceProfileRoleDetails_profileArn: str = None
    iamInstanceProfileRoleDetails_roleArn: str = None
    iamInstanceProfileRoleDetails_roleName: str = None
    # end of iamInstanceProfileRoleDetails fields
    imageId: str = None
    publicIpAddress: str = None
    instanceType: str = None
    ipAddress: str = None
    publicDnsName: str = None
    instanceStatus: str = None
    iamInstanceProfile: None = None  # parse out to below fields
    # iamInstanceProfile is parsed out to below fields
    iamInstanceProfile_name: str = None
    iamInstanceProfile_id: str = None
    iamInstanceProfile_arn: str = None
    # end of iamInstanceProfile fields
    privateIpAddress: str = None
    launchTime: Union[str, datetime] = None
    classifications: BaseList[str] = None  # ???
    spotInstanceRequestId: str = None
    elasticIpAddress: str = None
    secondaryPrivateIpAddress: str = None
    securityGroups: BaseList[str] = None  # list of dicts
    privateDnsName: str = None
    criticalityScore: int = None
    vulnerabilities: BaseList[
        str
    ] = None  # empty list unless resource details API is called

    def __post_init__(self):
        DICT_BL_FIELDS = ["networkInterfaceAddresses", "securityGroups"]
        STR_BL_FIELDS = ["events", "classifications"]
        PARSE_OUT_FIELDS = ["iamInstanceProfileRoleDetails", "iamInstanceProfile"]

        for field in DICT_BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    s = ""
                    for key in item.keys():
                        s += f"{key}: {item[key]}, "
                    s = f"({s[:-2]})"
                    bl.append(s)
                setattr(self, field, bl)

        for field in STR_BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    bl.append(item)
                setattr(self, field, bl)

        if self.vulnerabilityStats:
            data = self.vulnerabilityStats
            severity = data.get("severity", None)
            typeDetected = data.get("typeDetected", None)
            totalVulnerability = data.get("totalVulnerability", None)
            if severity:
                for key in severity.keys():
                    setattr(self, f"vulnerabilityStats_severity_{key}", severity[key])
            if typeDetected:
                for key in typeDetected.keys():
                    setattr(
                        self,
                        f"vulnerabilityStats_typeDetected_{key}",
                        typeDetected[key],
                    )
            if totalVulnerability:
                setattr(
                    self,
                    "vulnerabilityStats_typeDetected_totalVulnerability",
                    totalVulnerability,
                )

            setattr(self, "vulnerabilityStats", None)

        for field in PARSE_OUT_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                for key, value in data.items():
                    setattr(self, f"{field}_{key}", value)
                setattr(self, field, None)

        if self.launchTime:
            if not isinstance(self.launchTime, datetime):
                setattr(self, "launchTime", datetime.fromisoformat(self.launchTime))

        if self.vulnerabilities:
            data = self.vulnerabilities
            if isinstance(data, dict):
                data = [data]
            bl = BaseList()
            for vuln in data:
                bl.append(QID(**vuln))
            setattr(self, "vulnerabilities", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSEC2Instance object from a dictionary
        """
        return AWSEC2Instance(**data)


@dataclass
class AWSRouteTable(BaseResource):
    """
    Represents a Route Table in AWS
    """

    associations: BaseList[str] = None
    routeDtos: BaseList[str] = None
    routes: BaseList[str] = None
    vpcId: str = None
    routeTableId: str = None

    def __post_init__(self):
        BL_FIELDS = ["associations", "routeDtos", "routes"]

        for field in BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    s = ""
                    for key in item.keys():
                        s += f"{key}: {item[key]}, "
                    s = f"({s[:-2]})"
                    bl.append(s)
                setattr(self, field, bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSRouteTable object from a dictionary
        """
        return AWSRouteTable(**data)


@dataclass
class AWSEBSVolume(BaseResource):
    """
    Represents an EBS Volume in AWS
    """

    volumeType: str = None
    attachments: BaseList[str] = None
    availabilityZone: str = None
    volumeId: str = None
    iops: int = None
    state: str = None
    snapshotId: str = None
    encrypted: bool = None
    size: str = None
    createTime: Union[str, datetime] = None
    kmsKeyId: str = None

    def __post_init__(self):
        if self.attachments:
            data = self.attachments
            bl = BaseList()
            for attachment in data:
                s = ""
                for key in attachment.keys():
                    s += f"{key}: {attachment[key]}, "
                s = f"({s[:-2]})"
                bl.append(s)
            setattr(self, "attachments", bl)

        if self.createTime:
            if not isinstance(self.createTime, datetime):
                setattr(self, "createTime", datetime.fromisoformat(self.createTime))

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSEBSVolume object from a dictionary
        """
        return AWSEBSVolume(**data)


@dataclass
class AWSAutoScalingGroup(BaseResource):
    """
    Represents an Auto Scaling Group in AWS
    """

    healthCheckType: str = None
    instances: BaseList[str] = None  # list of strs
    autoScalingGroupARN: str = None
    availabilityZones: BaseList[str] = None  # list of strs
    launchConfigurationName: str = None
    autoScalingGroupName: str = None
    createdTime: Union[str, datetime] = None
    loadBalancerNames: BaseList[str] = None  # list of ???

    def __post_init__(self):
        STR_BL_FIELDS = [
            "instances",
            "availabilityZones",
            "loadBalancerNames",
            "loadBalancerNames",
        ]

        for field in STR_BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    bl.append(item)
                setattr(self, field, bl)

        if self.createdTime:
            if not isinstance(self.createdTime, datetime):
                setattr(self, "createdTime", datetime.fromisoformat(self.createdTime))

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSAutoScalingGroup object from a dictionary
        """
        return AWSAutoScalingGroup(**data)


@dataclass
class AWSEKSCluster(BaseResource):
    """
    Represents an EKS Cluster in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    associations: None = None  # parse out to below fields
    # associations is parsed out to below fields
    associations_resourcesVpcConfig_clusterSecurityGroupId: str = None
    assiciations_resourcesVpcConfig_securityGroupIds: BaseList[str] = None
    associations_resourcesVpcConfig_vpcId: str = None
    associations_resourcesVpcConfig_subnetIds: BaseList[str] = None
    version: float = None
    endpoint: str = None
    resourcesVpcConfig: None = None  # parse out to below fields
    # resourcesVpcConfig is parsed out to below fields
    resourcesVpcConfig_endpointPrivateAccess: bool = None
    resporcesVpcConfig_endpointPublicAccess: bool = None
    # end of resourcesVpcConfig fields
    roleArn: str = None
    identity: None = None  # parse out to below fields
    # identity is parsed out to below fields
    identity_oidc_issuer: str = None
    # end of identity fields
    platformVersion: str = None
    roleName: str = None
    logging: None = None  # parse out to below fields
    # logging is parsed out to below fields
    logging_clusterLogging: BaseList[str] = None
    # end of logging fields
    arn: str = None
    status: str = None

    def __post_init__(self):
        if self.version:
            setattr(self, "version", float(self.version))

        if self.associations:
            data = self.associations.get("resourcesVpcConfig", None)
            if data:
                for key, value in data.items():
                    if key == "securityGroupIds":
                        bl = BaseList()
                        for group in value:
                            bl.append(group)
                        setattr(
                            self, "associations_resourcesVpcConfig_securityGroupIds", bl
                        )
                    elif key == "subnetIds":
                        bl = BaseList()
                        for subnet in value:
                            bl.append(subnet)
                        setattr(self, "associations_resourcesVpcConfig_subnetIds", bl)
                    else:
                        setattr(self, f"associations_resourcesVpcConfig_{key}", value)
            setattr(self, "associations", None)

        if self.resourcesVpcConfig:
            data = self.resourcesVpcConfig
            for key, value in data.items():
                setattr(self, f"resourcesVpcConfig_{key}", value)
            setattr(self, "resourcesVpcConfig", None)

        if self.identity:
            data = self.identity.get("oidc", None)
            if data:
                data = data.get("issuer", None)
                if data:
                    setattr(self, "identity_oidc_issuer", data)
            setattr(self, "identity", None)

        if self.logging:
            data = self.logging.get("clusterLogging", None)[0]
            if data:
                bl = BaseList()
                i = data.get("types", None)
                for item in i:
                    bl.append(item)
                setattr(self, "logging_clusterLogging", bl)
            setattr(self, "logging", None)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSEKSCluster object from a dictionary
        """
        return AWSEKSCluster(**data)


@dataclass
class AWSEKSNodeGroup(BaseResource):
    """
    Represents an EKS Node Group in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    associations: None = None  # parse out to below fields
    # associations is parsed out to below fields
    associations_clusterArn: str = None
    associations_clusterName: str = None
    associations_subnets: BaseList[str] = None
    associations_resources_autoScalingGroups: BaseList[str] = None
    # end of associations fields
    nodeRoleName: str = None
    scalingConfig: None = None  # parse out to below fields
    # scalingConfig is parsed out to below fields
    scalingConfig_maxSize: int = None
    scalingConfig_desiredSize: int = None
    scalingConfig_minSize: int = None
    # end of scalingConfig fields
    capacityType: str = None
    instanceTypes: BaseList[str] = None
    releaseVersion: str = None
    version: float = None
    labels: BaseList[str] = None
    diskSize: int = None
    nodeRole: str = None
    launchTemplate: None = None  # parse out to below fields
    # launchTemplate is parsed out to below fields
    launchTemplate_name: str = None
    launchTemplate_id: str = None
    launchTemplate_version: str = None
    # end of launchTemplate fields
    amiType: str = None
    status: str = None
    health: BaseList[str] = None

    def __post_init__(self):
        if self.version:
            setattr(self, "version", float(self.version))

        if self.associations:
            data = self.associations
            setattr(self, "associations_clusterArn", data.get("clusterArn", None))
            setattr(self, "associations_clusterName", data.get("clusterName", None))
            subnets = data.get("subnets", None)
            if subnets:
                bl = BaseList()
                for subnet in subnets:
                    bl.append(subnet)
                setattr(self, "associations_subnets", bl)
            resources = data.get("resources", None)
            if resources:
                if resources.get("autoScalingGroups", None):
                    bl = BaseList()
                    for group in resources["autoScalingGroups"]:
                        bl.append(group.get("name", None))
                    setattr(self, "associations_resources_autoScalingGroups", bl)
            setattr(self, "associations", None)

        if self.scalingConfig:
            data = self.scalingConfig
            if data.get("maxSize", None):
                setattr(self, "scalingConfig_maxSize", int(data.get("maxSize", None)))
            if data.get("desiredSize", None):
                setattr(
                    self,
                    "scalingConfig_desiredSize",
                    int(data.get("desiredSize", None)),
                )
            if data.get("minSize", None):
                setattr(self, "scalingConfig_minSize", int(data.get("minSize", None)))
            setattr(self, "scalingConfig", None)

        if self.labels:
            data = self.labels
            bl = BaseList()
            for k, v in data.items():
                bl.append(f"({k}: {v})")

        if self.launchTemplate:
            data = self.launchTemplate
            setattr(self, "launchTemplate_name", data.get("name", None))
            setattr(self, "launchTemplate_id", data.get("id", None))
            setattr(self, "launchTemplate_version", data.get("version", None))
            setattr(self, "launchTemplate", None)

        if self.health:
            data = self.health.get("issues", None)
            if data:
                bl = BaseList()
                for issue in data:
                    code = issue.get("code", None)
                    message = issue.get("message", None)
                    resourceIds = issue.get("resourceIds", None)
                    resourcebl = BaseList()
                    if resourceIds:
                        for resourceId in resourceIds:
                            resourcebl.append(resourceId)
                    bl.append(
                        f"(code: {code}, message: {message}, resourceIds: {resourcebl})"
                    )

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSEKSNodeGroup object from a dictionary
        """
        return AWSEKSNodeGroup(**data)


@dataclass
class AWSEKSFargateProfile(BaseResource):
    """
    Represents an EKS Fargate Profile in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    associations: None = None  # parse out to below fields
    # associations is parsed out to below fields
    associations_clusterArn: str = None
    associations_clusterName: str = None
    associations_subnets: BaseList[str] = None
    # end of associations fields
    podExecutionRoleArn: str = None
    selectors: BaseList[str] = None
    podExecutionRoleName: str = None
    status: str = None

    def __post_init__(self):
        if self.associations:
            data = self.associations
            setattr(self, "associations_clusterArn", data.get("clusterArn", None))
            setattr(self, "associations_clusterName", data.get("clusterName", None))
            subnets = data.get("subnets", None)
            if subnets:
                bl = BaseList()
                for subnet in subnets:
                    bl.append(subnet)
                setattr(self, "associations_subnets", bl)
            setattr(self, "associations", None)

        if self.selectors:
            data = self.selectors
            bl = BaseList()
            for selector in data:
                s = ""
                namespace = selector.get("namespace", None)
                labels = selector.get("labels", None)
                if labels:
                    for k, v in labels.items():
                        s += f"{k}: {v}, "
                    # Take out the last comma and space
                    s = f"[{s[:-2]}]"

                bl.append(f"(namespace: {namespace}, labels: {s})")
            setattr(self, "selectors", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSEKSFargateProfile object from a dictionary
        """
        return AWSEKSFargateProfile(**data)


@dataclass
class AWSVPCEndpoint(BaseResource):
    """
    Represents a VPC Endpoint in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    policyDocument: str = None  # needs parsed with json.loads
    dnsOption: str = None
    serviceName: str = None
    securityGroupSet: BaseList[str] = None  # list of dicts
    privateDnsEnabled: bool = None
    requesterManaged: bool = None
    ipAddressType: str = None
    routeTableIds: BaseList[str] = None  # ???
    vpcEndpointType: str = None
    vpcId: str = None
    state: str = None
    dnsEntrySets: BaseList[str] = None  # list of dicts
    networkInterfaceIds: BaseList[str] = None  # list of strs
    subnetIds: BaseList[str] = None  # list of strs

    def __post_init__(self):
        STR_BL_FIELDS = ["routeTableIds", "networkInterfaceIds", "subnetIds"]

        for field in STR_BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    bl.append(item)
                setattr(self, field, bl)

        if self.policyDocument:
            data = loads(self.policyDocument).get("Statement", None)
            if data:
                bl = BaseList()
                for statement in data:
                    s = ""
                    for key in statement.keys():
                        s += f"{key}: {statement[key]}, "
                    s = f"({s[:-2]})"
                    bl.append(s)
                setattr(self, "policyDocument", bl)

        if self.securityGroupSet:
            data = self.securityGroupSet
            bl = BaseList()
            for group in data:
                groupName = group.get("groupName", None)
                groupId = group.get("groupId", None)
                bl.append(f"(groupName: {groupName}, groupId: {groupId})")
            setattr(self, "securityGroupSet", bl)

        if self.dnsEntrySets:
            data = self.dnsEntrySets
            bl = BaseList()
            for entry in data:
                hostedZoneId = entry.get("hostedZoneId", None)
                dnsName = entry.get("dnsName", None)
                bl.append(f"(hostedZoneId: {hostedZoneId}, dnsName: {dnsName})")
            setattr(self, "dnsEntrySets", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSVPCEndpoint object from a dictionary
        """
        return AWSVPCEndpoint(**data)


@dataclass
class AWSVPCEndpointService(BaseResource):
    """
    Represents a VPC Endpoint Service in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    owner: str = None
    acceptanceRequired: bool = None
    managesVpcEndpoints: bool = None
    availabilityZone: BaseList[str] = None  # list of strs
    baseEndpointDnsName: BaseList[str] = None  # list of strs
    supportedIpAddressType: str = None
    serviceTypes: str = None
    privateDnsName: str = None
    privateDnsNameVerificationState: str = None
    vpcEndpointPolicySupported: bool = None

    def __post_init__(self):
        STR_BL_FIELDS = ["availabilityZone", "baseEndpointDnsName"]

        for field in STR_BL_FIELDS:
            if getattr(self, field):
                data = getattr(self, field)
                bl = BaseList()
                for item in data:
                    bl.append(item)
                setattr(self, field, bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSVPCEndpointService object from a dictionary
        """
        return AWSVPCEndpointService(**data)


@dataclass
class AWSIAMGroup(BaseResource):
    """
    Represents an IAM Group in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    classifications: BaseList[str] = None  # list of strs
    groupId: str = None
    GroupPolicyList: BaseList[str] = None  # list of dicts
    AttachedManagedPolicies: BaseList[str] = None  # list of dicts
    arn: str = None

    def __post_init__(self):
        if self.classifications:
            data = self.classifications
            bl = BaseList()
            for classification in data:
                bl.append(classification)

        if self.GroupPolicyList:
            data = self.GroupPolicyList
            bl = BaseList()
            for policy in data:
                policyName = policy.get("PolicyName", None)
                policyDocument = loads(policy.get("PolicyDocument", None))
                if policyDocument:
                    policyDocument_version = policyDocument.get("Version", None)
                    policyDocument_statement = policyDocument.get("Statement", None)
                    if policyDocument_statement:
                        statementbl = BaseList()
                        if isinstance(policyDocument_statement, list):
                            for statement in policyDocument_statement:
                                action = statement.get("Action", None)
                                resource = statement.get("Resource", None)
                                effect = statement.get("Effect", None)
                                statementbl.append(
                                    f"(Action: {action}, Resource: {resource}, Effect: {effect})"
                                )
                        else:
                            action = policyDocument_statement.get("Action", None)
                            resource = policyDocument_statement.get("Resource", None)
                            effect = policyDocument_statement.get("Effect", None)
                            statementbl.append(
                                f"(Action: {action}, Resource: {resource}, Effect: {effect})"
                            )

                        policyDocument = f"(Version: {policyDocument_version}, Statement: {statementbl})"
                bl.append(
                    f"(policyName: {policyName}, policyDocument: {policyDocument})"
                )
            setattr(self, "GroupPolicyList", bl)

        if self.AttachedManagedPolicies:
            data = self.AttachedManagedPolicies
            bl = BaseList()
            for policy in data:
                policyArn = policy.get("PolicyArn", None)
                policyName = policy.get("PolicyName", None)
                bl.append(f"(PolicyArn: {policyArn}, PolicyName: {policyName})")
            setattr(self, "AttachedManagedPolicies", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSIAMGroup object from a dictionary
        """
        return AWSIAMGroup(**data)


@dataclass
class AWSIAMPolicy(BaseResource):
    """
    Represents an IAM Policy in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    IsAttachable: bool = None
    PermissionsBoundaryUsageCount: int = None
    AttachmentCount: int = None
    DefaultVersionId: str = None
    defaultPolicyVersion: None = None  # parse out to below fields
    # defaultPolicyVersion is parsed out to below fields
    defaultPolicyVersion_VersionId: str = None
    defaultPolicyVersion_IsDefaultVersion: bool = None
    defaultPolicyVersion_Document_Version: str = None
    defaultPolicyVersion_Document_Statement: BaseList[str] = None  # list of dicts
    defaultPolicyVersion_CreateDate: Union[str, datetime] = None
    # end of defaultPolicyVersion fields
    _type: str = None
    path: str = None
    classifications: BaseList[str] = None  # list of strs
    policyId: str = None
    arn: str = None

    def __post_init__(self):
        if self.defaultPolicyVersion:
            data = self.defaultPolicyVersion
            setattr(self, "defaultPolicyVersion_VersionId", data.get("VersionId", None))
            setattr(
                self,
                "defaultPolicyVersion_IsDefaultVersion",
                data.get("IsDefaultVersion", None),
            )
            createDate = data.get("CreateDate", None)
            if createDate:
                setattr(
                    self,
                    "defaultPolicyVersion_CreateDate",
                    datetime.fromtimestamp(createDate),
                )
            document = data.get("Document", None)
            if document:
                document = loads(document)
                setattr(
                    self,
                    "defaultPolicyVersion_Document_Version",
                    document.get("Version", None),
                )
                statement = document.get("Statement", None)
                if statement:
                    bl = BaseList()
                    if isinstance(statement, list):
                        for stmt in statement:
                            resource = stmt.get("Resource", None)
                            stmt.pop("Resource", None)
                            effect = stmt.get("Effect", None)
                            stmt.pop("Effect", None)
                            # For the remaining key (either Action or Condition),
                            # it is a list of strings. We need to convert to BaseList
                            # for consistency
                            is_action = True if "Action" in stmt else False
                            rule = stmt.get("Action", None)
                            stmt.pop("Action", None)
                            if not rule:
                                rule = stmt.get("Condition", None)
                                stmt.pop("Condition", None)
                            if rule:
                                rulebl = BaseList()
                                for r in rule:
                                    rulebl.append(r)
                                rule = rulebl
                    else:
                        resource = statement.get("Resource", None)
                        statement.pop("Resource", None)
                        effect = statement.get("Effect", None)
                        statement.pop("Effect", None)
                        is_action = True if "Action" in statement else False
                        rule = statement.get("Action", None)
                        statement.pop("Action", None)
                        if not rule:
                            rule = statement.get("Condition", None)
                            statement.pop("Condition", None)
                        if rule:
                            rulebl = BaseList()
                            for r in rule:
                                rulebl.append(r)
                            rule = rulebl

                        # Now we have the resource, effect, and rule
                        # Add them to the statement
                        s = f"(Resource: {resource}, Effect: {effect}, {'Action' if is_action else 'Condition'}: {str(rule)})"
                        bl.append(s)
                    setattr(self, "defaultPolicyVersion_Document_Statement", bl)
            setattr(self, "defaultPolicyVersion", None)

        if self.classifications:
            data = self.classifications
            bl = BaseList()
            for classification in data:
                bl.append(classification)
            setattr(self, "classifications", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSIAMPolicy object from a dictionary
        """
        return AWSIAMPolicy(**data)


@dataclass
class AWSIAMRole(BaseResource):
    """
    Represents an IAM Role in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    AttachedManagedPolicies: BaseList[str] = None  # list of dicts
    RoleLastUsed: str = None
    arn: str = None
    roleId: str = None
    PermissionsBoundary: None = None  # parse out to below fields
    # PermissionsBoundary is parsed out to below fields
    PermissionsBoundary_PermissionsBoundaryArn: str = None
    PermissionsBoundary_PermissionsBoundaryType: str = None
    # end of PermissionsBoundary fields
    path: str = None
    classifications: BaseList[str] = None  # list of strs
    AssumeRolePolicyDocument: None = None
    # AssumeRolePolicyDocument is parsed out to below fields
    AssumeRolePolicyDocument_Version: str = None
    AssumeRolePolicyDocument_Statement: BaseList[str] = None
    # end of AssumeRolePolicyDocument fields
    trustedEntities: BaseList[str] = None  # list of strs {'AWS Service': ['ops']}
    RolePolicyList: BaseList[str] = None  # list of dicts
    InstanceProfileList: BaseList[str] = None  # list of dicts

    def __post_init__(self):
        if self.AttachedManagedPolicies:
            data = self.AttachedManagedPolicies
            bl = BaseList()
            for policy in data:
                policyArn = policy.get("PolicyArn", None)
                policyName = policy.get("PolicyName", None)
                bl.append(f"(PolicyArn: {policyArn}, PolicyName: {policyName})")
            setattr(self, "AttachedManagedPolicies", bl)

        if self.PermissionsBoundary:
            data = self.PermissionsBoundary
            setattr(
                self,
                "PermissionsBoundary_PermissionsBoundaryArn",
                data.get("PermissionsBoundaryArn", None),
            )
            setattr(
                self,
                "PermissionsBoundary_PermissionsBoundaryType",
                data.get("PermissionsBoundaryType", None),
            )
            setattr(self, "PermissionsBoundary", None)

        if self.classifications:
            data = self.classifications
            bl = BaseList()
            for classification in data:
                bl.append(classification)
            setattr(self, "classifications", bl)

        if self.AssumeRolePolicyDocument:
            data = loads(self.AssumeRolePolicyDocument)
            setattr(self, "AssumeRolePolicyDocument_Version", data.get("Version", None))
            statement = data.get("Statement", None)
            if statement:
                bl = BaseList()
                for stmt in statement:
                    effect = stmt.get("Effect", None)
                    stmt.pop("Effect", None)
                    principal = stmt.get("Principal", None)
                    stmt.pop("Principal", None)
                    action = stmt.get("Action", None)
                    stmt.pop("Action", None)
                    resource = stmt.get("Resource", None)
                    stmt.pop("Resource", None)
                    s = f"(Effect: {effect}, Principal: {principal}, Action: {action}, Resource: {resource})"
                    bl.append(s)
                setattr(self, "AssumeRolePolicyDocument_Statement", bl)
            setattr(self, "AssumeRolePolicyDocument", None)

        if self.trustedEntities:
            data = self.trustedEntities
            for k, v in data.items():
                formatted = f"{k}: {v}"
            setattr(self, "trustedEntities", formatted)

        if self.RolePolicyList:
            data = self.RolePolicyList
            bl = BaseList()
            for policy in data:
                policyName = policy.get("PolicyName", None)
                policyDocument = loads(policy.get("PolicyDocument", None))
                if policyDocument:
                    policyDocument_version = policyDocument.get("Version", None)
                    policyDocument_statement = policyDocument.get("Statement", None)
                    if policyDocument_statement:
                        if isinstance(policyDocument_statement, list):
                            statementbl = BaseList()
                            for statement in policyDocument_statement:
                                action = statement.get("Action", None)
                                resource = statement.get("Resource", None)
                                effect = statement.get("Effect", None)
                                statementbl.append(
                                    f"(Action: {action}, Resource: {resource}, Effect: {effect})"
                                )
                            policyDocument = f"(Version: {policyDocument_version}, Statement: {statementbl})"
                        else:
                            policyDocument = f"(Version: {policyDocument_version}, Statement: {policyDocument_statement})"
                bl.append(
                    f"(policyName: {policyName}, policyDocument: {policyDocument})"
                )
            setattr(self, "RolePolicyList", bl)

        if self.InstanceProfileList:
            data = self.InstanceProfileList
            bl = BaseList()
            for profile in data:
                path = profile.get("Path", None)
                instanceProfileName = profile.get("InstanceProfileName", None)
                InstanceProfileId = profile.get("InstanceProfileId", None)
                arn = profile.get("Arn", None)
                CreateDate = profile.get("CreateDate", None)  # timestamp
                if CreateDate and not isinstance(CreateDate, datetime):
                    CreateDate = datetime.fromtimestamp(CreateDate)
                roles = profile.get("Roles", None)
                if roles:
                    rolebl = BaseList()
                    for role in roles:
                        rolepath = role.get("Path", None)
                        rolename = role.get("RoleName", None)
                        roleassumeRolePolicyDocument = role.get(
                            "AssumeRolePolicyDocument", None
                        )
                        if roleassumeRolePolicyDocument:
                            roleassumeRolePolicyDocument = unquote(
                                roleassumeRolePolicyDocument
                            )
                        rolearn = role.get("Arn", None)
                        rolecreateDate = role.get("CreateDate", None)
                        if rolecreateDate and not isinstance(rolecreateDate, datetime):
                            rolecreateDate = datetime.fromtimestamp(rolecreateDate)
                        roleId = role.get("RoleId", None)
                        rolestr = f"(Path: {rolepath}, RoleName: {rolename}, Arn: {rolearn}, CreateDate: {rolecreateDate}, RoleId: {roleId}, AssumeRolePolicyDocument: {roleassumeRolePolicyDocument})"
                        rolebl.append(rolestr)
                    roles = rolebl
                s = f"(Path: {path}, InstanceProfileName: {instanceProfileName}, InstanceProfileId: {InstanceProfileId}, Arn: {arn}, CreateDate: {CreateDate}, Roles: {roles})"
                bl.append(s)
            setattr(self, "InstanceProfileList", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSIAMRole object from a dictionary
        """
        return AWSIAMRole(**data)


@dataclass
class AWSSagemakerNotebook(BaseResource):
    """
    Represents a SageMaker Notebook in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    subnetId: str = None
    networkInterfaceId: str = None
    volumeSizeInGB: float = None
    rootAccess: bool = None
    instanceType: str = None
    platformId: str = None
    url: str = None
    directInternetAccess: str = None
    roleArn: str = None
    securityGroups: BaseList[str] = None  # list of strs
    status: str = None
    defaultCodeRepository: str = None
    kmsKeyId: str = None

    def __post_init__(self):
        if self.securityGroups:
            data = self.securityGroups
            bl = BaseList()
            for group in data:
                bl.append(group)
            setattr(self, "securityGroups", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSSagemakerNotebook object from a dictionary
        """
        return AWSSagemakerNotebook(**data)


@dataclass
class AWSCloudfrontDistribution(BaseResource):
    """
    Represents a Cloudfront Distribution in AWS
    """

    connectorId: str = None
    customerId: str = None
    scanId: str = None
    displayName: str = None
    defaultCacheBehavior: None = None  # parse out to below fields
    defaultCacheBehavior_Compress: bool = None
    defaultCacheBehavior_FunctionAssociations_Quantity: int = None
    defaultCacheBehavior_FunctionAssociations_Items: BaseList[str] = None
    defaultCacheBehavior_LambdaFunctionAssociations_Quantity: int = None
    defaultCacheBehavior_LambdaFunctionAssociations_Items: BaseList[str] = None
    defaultCacheBehavior_TargetOriginId: str = None
    defaultCacheBehavior_ViewerProtocolPolicy: str = None
    defaultCacheBehavior_TrustedSigners_Enabled: bool = (
        None  # parsed out of trustedsigners dict
    )
    defaultCacheBehavior_TrustedSigners_Quantity: int = None  # ^^^
    defaultCacheBehavior_FieldLevelEncryptionId: str = None
    defaultCacheBehavior_DefaultTTL: int = None
    defaultCacheBehavior_TrustedKeyGroups_Enabled: bool = (
        None  # parsed out of trustedkeygroups dict
    )
    defaultCacheBehavior_TrustedKeyGroups_Quantity: int = None  # ^^^
    defaultCacheBehavior_AllowedMethods_CachedMethods_Quantity: int = (
        None  # parsed out of allowedmethods dict
    )
    defaultCacheBehavior_AllowedMethods_CachedMethods_Items: BaseList[str] = None  # ^^^
    defaultCacheBehavior_AllowedMethods_Items_Method: BaseList[str] = None  # ^^^
    defaultCacheBehavior_SmoothStreaming: bool = None
    defaultCacheBehavior_MinTTL: int = None
    defaultCacheBehavior_MaxTTL: int = None
    # end of defaultCacheBehavior fields
    loggingEnabled: bool = None
    isIpv6Enabled: bool = None
    includeCookies: bool = None
    geoRestriction: None = None  # parse out to below fields
    geoRestriction_RestrictionType: str = None  # parsed out of georestriction dict
    geoRestriction_Quantity: int = None  # ^^^
    # end of geoRestriction fields
    enabled: bool = None
    webAclId: str = None
    minimumProtocolVersion: str = None
    httpVersion: str = None
    priceClass: str = None
    loggingBucket: str = None
    origins: BaseList[str] = None  # list of dicts
    comment: str = None
    id: str = None
    arn: str = None
    staging: bool = None
    status: str = None
    loggingBucketPrefix: str = None
    acmCertificateARN: str = None

    def __post_init__(self):
        if self.defaultCacheBehavior:
            # Sigh...
            data = self.defaultCacheBehavior
            setattr(self, "defaultCacheBehavior_Compress", data.get("Compress", None))
            setattr(
                self,
                "defaultCacheBehavior_TargetOriginId",
                data.get("TargetOriginId", None),
            )
            setattr(
                self,
                "defaultCacheBehavior_ViewerProtocolPolicy",
                data.get("ViewerProtocolPolicy", None),
            )
            if data.get("TrustedSigners", None):
                setattr(
                    self,
                    "defaultCacheBehavior_TrustedSigners_Enabled",
                    data["TrustedSigners"].get("Enabled", None),
                )
                setattr(
                    self,
                    "defaultCacheBehavior_TrustedSigners_Quantity",
                    data["TrustedSigners"].get("Quantity", None),
                )
            setattr(
                self,
                "defaultCacheBehavior_FieldLevelEncryptionId",
                data.get("FieldLevelEncryptionId", None),
            )
            setattr(
                self, "defaultCacheBehavior_DefaultTTL", data.get("DefaultTTL", None)
            )
            if data.get("TrustedKeyGroups", None):
                setattr(
                    self,
                    "defaultCacheBehavior_TrustedKeyGroups_Enabled",
                    data["TrustedKeyGroups"].get("Enabled", None),
                )
                setattr(
                    self,
                    "defaultCacheBehavior_TrustedKeyGroups_Quantity",
                    data["TrustedKeyGroups"].get("Quantity", None),
                )

            if data.get("FunctionAssociations", None):
                setattr(
                    self,
                    "defaultCacheBehavior_FunctionAssociations_Quantity",
                    data["FunctionAssociations"].get("Quantity", None),
                )
                if data["FunctionAssociations"].get("Items", None):
                    bl = BaseList()
                    for item in data["FunctionAssociations"]["Items"]:
                        bl.append(item)
                    setattr(self, "defaultCacheBehavior_LambdaFunctionAssociations", bl)

            if data.get("LambdaFunctionAssociations", None):
                setattr(
                    self,
                    "defaultCacheBehavior_LambdaFunctionAssociations_Quantity",
                    data["LambdaFunctionAssociations"].get("Quantity", None),
                )
                if data["LambdaFunctionAssociations"].get("Items", None):
                    bl = BaseList()
                    for item in data["LambdaFunctionAssociations"]["Items"]:
                        bl.append(item)
                    setattr(self, "defaultCacheBehavior_LambdaFunctionAssociations", bl)

            if data.get("AllowedMethods", None):
                if data["AllowedMethods"].get("CachedMethods", None):
                    setattr(
                        self,
                        "defaultCacheBehavior_AllowedMethods_CachedMethods_Quantity",
                        data["AllowedMethods"]["CachedMethods"].get("Quantity", None),
                    )
                    if data["AllowedMethods"]["CachedMethods"].get("Items", None):
                        bl = BaseList()
                        for item in data["AllowedMethods"]["CachedMethods"]["Items"][
                            "Method"
                        ]:
                            bl.append(item)
                        setattr(
                            self,
                            "defaultCacheBehavior_AllowedMethods_CachedMethods_Items",
                            bl,
                        )
                if data["AllowedMethods"].get("Items", None):
                    bl = BaseList()
                    for item in data["AllowedMethods"]["Items"]["Method"]:
                        bl.append(item)
                    setattr(
                        self, "defaultCacheBehavior_AllowedMethods_Items_Method", bl
                    )

            setattr(
                self,
                "defaultCacheBehavior_SmoothStreaming",
                data.get("SmoothStreaming", None),
            )
            setattr(self, "defaultCacheBehavior_MinTTL", data.get("MinTTL", None))
            setattr(self, "defaultCacheBehavior_MaxTTL", data.get("MaxTTL", None))
            setattr(self, "defaultCacheBehavior", None)

        if self.geoRestriction:
            restriction = (
                self,
                "geoRestriction_RestrictionType",
                self.geoRestriction.get("RestrictionType", None),
            )
            if restriction == "none":
                restriction = None
            setattr(self, "geoRestriction_RestrictionType", restriction)
            setattr(self, "geoRestriction", None)

        if self.origins:
            data = self.origins
            bl = BaseList()
            for origin in data:
                if isinstance(origin, list):
                    for o in origin:
                        connectionTimeout = o.get("ConnectionTimeout", None)
                        originAccessControlId = o.get("OriginAccessControlId", None)
                        connectionAttempts = o.get("ConnectionAttempts", None)
                        domainName = o.get("DomainName", None)
                        originPath = o.get("OriginPath", None)
                        Id = o.get("Id", None)
                        s = f"(ConnectionTimeout: {connectionTimeout}, OriginAccessControlId: {originAccessControlId}, ConnectionAttempts: {connectionAttempts}, DomainName: {domainName}, OriginPath: {originPath}, Id: {Id})"
                        bl.append(s)
                else:
                    connectionTimeout = origin.get("ConnectionTimeout", None)
                    originAccessControlId = origin.get("OriginAccessControlId", None)
                    connectionAttempts = origin.get("ConnectionAttempts", None)
                    domainName = origin.get("DomainName", None)
                    originPath = origin.get("OriginPath", None)
                    Id = origin.get("Id", None)
                    s = f"(ConnectionTimeout: {connectionTimeout}, OriginAccessControlId: {originAccessControlId}, ConnectionAttempts: {connectionAttempts}, DomainName: {domainName}, OriginPath: {originPath}, Id: {Id})"
                    bl.append(s)
            setattr(self, "origins", bl)

        super().__post_init__()

    @staticmethod
    def from_dict(data):
        """
        from_dict - create an AWSCloudfrontDistribution object from a dictionary
        """
        return AWSCloudfrontDistribution(**data)



"""
software.py - contains the Software dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class Software:
    """
    Software - represents a single software entry in a SoftwareList.

    This class is used to represent a single software entry in a SoftwareList,
    which is used to represent the software that is affected by a vulnerability.
    """

    PRODUCT: str = field(metadata={"description": "The name of the software."})
    VENDOR: str = field(
        metadata={"description": "The vendor of the software."},
        default="",
        compare=False,
    )

    def __str__(self):
        return self.PRODUCT

    def __contains__(
        self, item
    ):  # allows us to use the 'in' operator. for example, 'if "Adobe" in software'. this is a fuzzy search.
        # see if it was found in the name or vendor:
        return item in self.PRODUCT or item in self.VENDOR

    def copy(self):
        return Software(PRODUCT=self.PRODUCT, VENDOR=self.VENDOR)

    def is_vendor(self, vendor: str):
        return self.VENDOR.lower() == vendor.lower()

    def is_name(self, product: str):
        return self.PRODUCT.lower() == product.lower()

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Union[dict, list]):
        """
        from_dict - create a Software object from a dictionary.

        This function is used to create a Software object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"PRODUCT", "VENDOR"}

        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)
